import json
import numpy as np
import skopt

from suggestion.models import Study
from suggestion.models import Trial
from suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm


class BaseSkoptAlgorithm(AbstractSuggestionAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    Refer to https://github.com/scikit-optimize/scikit-optimize .
    """

    def __init__(self, algorithm_name="bayesian_optimization"):

        self.algorithm_name = algorithm_name

    def get_new_suggestions(self, study_name, input_trials=[], number=1):
        """
        Get the new suggested trials with skopt algorithm.
        """

        # Construct search space, example: {"x": hyperopt.hp.uniform('x', -10, 10), "x2": hyperopt.hp.uniform('x2', -10, 10)}
        hyperopt_search_space = {}

        study = Study.objects.get(name=study_name)
        study_configuration_json = json.loads(study.study_configuration)
        params = study_configuration_json["params"]

        skopt_search_space = []

        for param in params:
            param_name = param["parameterName"]

            if param["type"] == "INTEGER":
                skopt_search_space.append(
                    skopt.space.Integer(
                        param["minValue"], param["maxValue"], name="min_samples_leaf"
                    )
                )

            elif param["type"] == "DOUBLE":
                skopt_search_space.append(
                    skopt.space.Real(
                        param["minValue"],
                        param["maxValue"],
                        "log-uniform",
                        name="learning_rate",
                    )
                )

            elif param["type"] == "DISCRETE" or param["type"] == "CATEGORICAL":
                pass

        if self.algorithm_name == "bayesian_optimization":
            skopt_optimizer = skopt.Optimizer([(-2.0, 2.0)])
        else:
            print("Unsupport skopt algorithm: {}".format(self.algorithm_name))

        completed_advisor_trials = Trial.objects.filter(
            study_name=study_name, status="Completed"
        )

        for index, advisor_trial in enumerate(completed_advisor_trials):
            # Example: {"learning_rate": 0.01, "optimizer": "ftrl"}
            parameter_values_json = json.loads(advisor_trial.parameter_values)

            # Example: [(-2.0, 2.0)]
            skopt_suggested = []

            for param in params:

                if param["type"] == "INTEGER" or param["type"] == "DOUBLE":
                    parameter_value = parameter_values_json[param["parameterName"]]
                    skopt_suggested.append(parameter_value)

                elif param["type"] == "DISCRETE":
                    pass

                elif param["type"] == "CATEGORICAL":
                    pass

            loss_for_skopt = advisor_trial.objective_value
            if study_configuration_json["goal"] == "MAXIMIZE":
                # Now hyperopt only supports fmin and we need to reverse objective value for maximization
                loss_for_skopt = -1 * advisor_trial.objective_value

            skopt_optimizer.tell(skopt_suggested, loss_for_skopt)

        return_trial_list = []

        for i in range(number):

            skopt_suggested = skopt_optimizer.ask()

            new_advisor_trial = Trial.create(study.name, "SkoptTrial")
            parameter_values_json = {}

            index = 0
            for param in params:

                if param["type"] == "INTEGER" or param["type"] == "DOUBLE":
                    parameter_values_json[param["parameterName"]] = skopt_suggested[
                        index
                    ]

                elif param["type"] == "DISCRETE":
                    pass

                elif param["type"] == "CATEGORICAL":
                    pass

                index += 1

            new_advisor_trial.parameter_values = json.dumps(parameter_values_json)
            new_advisor_trial.save()
            return_trial_list.append(new_advisor_trial)

        return return_trial_list



import json

from suggestion.models import Study
from suggestion.models import TrialMetric
from suggestion.early_stop_algorithm.abstract_early_stop import (
    AbstractEarlyStopAlgorithm,
)


class EarlyStopDescendingAlgorithm(AbstractEarlyStopAlgorithm):
    def get_early_stop_trials(self, trials):
        result = []

        for trial in trials:
            study = Study.objects.get(name=trial.study_name)
            study_configuration_json = json.loads(study.study_configuration)
            study_goal = study_configuration_json["goal"]

            metrics = TrialMetric.objects.filter(trial_id=trial.id).order_by(
                "-training_step"
            )
            metrics = [metric for metric in metrics]

            if len(metrics) >= 2:
                if study_goal == "MAXIMIZE":
                    if metrics[0].objective_value < metrics[1].objective_value:
                        result.append(trial)
                elif study_goal == "MINIMIZE":
                    if metrics[0].objective_value > metrics[1].objective_value:
                        result.append(trial)

        return result



import hyperopt

from . import register_hpo
from .advisorbase import AdvisorBaseHPOptimizer
from .suggestion.algorithm.mocmaes import MocmaesAlgorithm


@register_hpo("mocmaes")
class MocmaesAdvisorChoco(AdvisorBaseHPOptimizer):
    """
    MOCMAES algorithm in `advisor` package
    See https://github.com/tobegit3hub/advisor for the package
    See .advisorbase.AdvisorBaseHPOptimizer for more information
    """

    def __init__(self, args):
        super().__init__(args)
        self.method = MocmaesAlgorithm()



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch_geometric.utils import add_self_loops, remove_self_loops
# from operations import *

from .op_graph_classification import *
from torch.autograd import Variable
from .genotypes import NA_PRIMITIVES, LA_PRIMITIVES, POOL_PRIMITIVES, READOUT_PRIMITIVES, ACT_PRIMITIVES
# from genotypes import Genotype
from torch_geometric.nn import  global_mean_pool, global_add_pool
from .pooling_zoo import filter_features, filter_perm
from .archgen import AG
from .encoder import GEncoder
def act_map(act):
    if act == "linear":
        return lambda x: x
    if act == "elu":
        return torch.nn.ELU
    elif act == "sigmoid":
        return torch.nn.Sigmoid
    elif act == "tanh":
        return torch.nn.Tanh
    elif act == "relu":
        return torch.nn.ReLU
    elif act == "relu6":
        return torch.nn.ReLU6
    elif act == "softplus":
        return torch.nn.Softplus
    elif act == "leaky_relu":
        return torch.nn.LeakyReLU
    else:
        raise Exception("wrong activate function")

class NaMixedOp(nn.Module):

  def __init__(self, in_dim, out_dim, with_linear):
    super(NaMixedOp, self).__init__()
    self._ops = nn.ModuleList()

    for primitive in NA_PRIMITIVES:
      op = NA_OPS[primitive](in_dim, out_dim)
      self._ops.append(op)

      if with_linear:
        self._ops_linear = nn.ModuleList()
        op_linear = torch.nn.Linear(in_dim, out_dim)
        self._ops_linear.append(op_linear)

      # self.act = act_map(act)

  def forward(self, x, weights, edge_index, edge_weights, edge_attr, with_linear, graph_match = False):
    mixed_res = []
    if with_linear:
      for op, linear in zip(self._ops, self._ops_linear):
        mixed_res.append(op(x, edge_index, edge_weight=edge_weights, edge_attr=edge_attr)+linear(x))
        # print('with linear')
    else:
      for op in self._ops:
        mixed_res.append(op(x, edge_index, edge_weight=edge_weights, edge_attr=edge_attr))
        # print('without linear')

    if graph_match:
        # weights: batch * op
        opmap = torch.stack(mixed_res, 1) # batch * op * dim
        weights = weights.unsqueeze(1) # batch * 1 * op
        res = weights @ opmap # batch * 1 * dim
        res = res.squeeze() # batch * dim
        return res

    else:
        return sum([w * op for w, op in zip(weights, mixed_res)])

class LaMixedOp(nn.Module):

  def __init__(self, hidden_size, num_layers=None):
    super(LaMixedOp, self).__init__()
    self._ops = nn.ModuleList()
    for primitive in LA_PRIMITIVES:
      op = LA_OPS[primitive](hidden_size, num_layers)
      self._ops.append(op)

  def forward(self, x, weights):
    mixed_res = []
    for w, op in zip(weights, self._ops):
      # mixed_res.append(w * F.relu(op(x)))
      mixed_res.append(w * F.elu(op(x)))
    return sum(mixed_res)

def index_to_mask(index, size):
    mask = torch.zeros(size, dtype=torch.float64, device=index.device)
    new_index = index.fill_(index[0]).type(torch.long)
    mask[new_index] = 1.0
    return mask

class PoolingMixedOp(nn.Module):
    def __init__(self, hidden, ratio, num_nodes=0):
        super(PoolingMixedOp, self).__init__()
        self._ops = nn.ModuleList()
        for primitive in POOL_PRIMITIVES:
            op = POOL_OPS[primitive](hidden, ratio, num_nodes)
            self._ops.append(op)

    def forward(self, x, edge_index, edge_weights, data, batch, mask, weights):
        new_x = []
        new_edge_weight = []
        new_perm = []
        # neither add or ewmove self_loop, so edge_index remain unchanged.
        for w, op in zip(weights, self._ops):
            # mixed_res.append(w * F.relu(op(x)))
            x_tmp, edge_index, edge_weight_tmp, batch, perm = op(x, edge_index, edge_weights, data, batch, mask)
            #print(perm.size(), w)
            new_x.append(x_tmp * w)
            new_edge_weight.append(w * edge_weight_tmp)
            new_perm.append(w * index_to_mask(perm, x.size(0)))

        #remove nodes with perm
        x, edge_index, edge_weight, batch, perm = filter_perm(sum(new_x), edge_index, sum(new_edge_weight), batch, sum(new_perm), th=0.01)
        return x, edge_index, edge_weight, batch, perm

class ReadoutMixedOp(nn.Module):
    def __init__(self, hidden):
        super(ReadoutMixedOp, self).__init__()
        self._ops = nn.ModuleList()
        for primitive in READOUT_PRIMITIVES:
            op = READOUT_OPS[primitive](hidden)
            self._ops.append(op)

    def forward(self, x, batch, mask, weights):
        mixed_res = []
        for w, op in zip(weights, self._ops):
            tmp_res = w * op(x, batch, mask)
            # print('readout', tmp_res.size())
            mixed_res.append(tmp_res)
        return sum(mixed_res)

class ActMixedOp(nn.Module):
    def __init__(self):
        super(ActMixedOp, self).__init__()
        self._ops = nn.ModuleDict()
        for primitive in ACT_PRIMITIVES:
            if primitive == 'linear':
                self._ops[primitive] = act_map(primitive)
            else:
                self._ops[primitive] = act_map(primitive)()

    def forward(self, x,  weights):
        mixed_res = []

        for i in range(len(ACT_PRIMITIVES)):
            mixed_res.append(weights[i] * self._ops[ACT_PRIMITIVES[i]](x))
        return sum(mixed_res)

class Network(nn.Module):

  def __init__(self, criterion, in_dim, out_dim, hidden_size, num_layers=3, dropout=0.5, epsilon=0.0, args=None, with_conv_linear=False,num_nodes=0, mol=False, virtual=False):
    super(Network, self).__init__()

    self.in_dim = in_dim
    self.out_dim = out_dim
    self.hidden_size = hidden_size
    self.num_layers = num_layers
    self.num_nodes = num_nodes
    self._criterion = criterion
    self.dropout = dropout
    self.epsilon = epsilon
    self.with_linear = with_conv_linear
    self.explore_num = 0
    self.args = args
    self.temp = args.temp
    self._loc_mean = args.loc_mean
    self._loc_std = args.loc_std
    self.mol = mol # if the task is molecule
    self.virtual = virtual

    if num_layers == 1:
        self.pooling_ratio = [0.1]
    elif num_layers == 2:
        self.pooling_ratio = [0.25, 0.25]
    elif num_layers == 3:
        self.pooling_ratio = [0.5, 0.5, 0.5]
    elif num_layers == 4:
        self.pooling_ratio = [0.6, 0.6, 0.6, 0.6]
    elif num_layers == 5:
        self.pooling_ratio = [0.7, 0.7, 0.7, 0.7, 0.7]
    elif num_layers == 6:
        self.pooling_ratio = [0.8, 0.8, 0.8, 0.8, 0.8, 0.8]

    self.lin1 = nn.Linear(in_dim, hidden_size)
    self.atom_encoder = AtomEncoder(hidden_size)
    self.virtualnode_embedding = torch.nn.Embedding(1, hidden_size)
    torch.nn.init.constant_(self.virtualnode_embedding.weight.data, 0)

    self.mlp_virtualnode_list = torch.nn.ModuleList()
    for layer in range(num_layers - 1):
        self.mlp_virtualnode_list.append(torch.nn.Sequential(torch.nn.Linear(hidden_size, 2*hidden_size), torch.nn.BatchNorm1d(2*hidden_size), torch.nn.ReLU(), \
                                                torch.nn.Linear(2*hidden_size, hidden_size), torch.nn.BatchNorm1d(hidden_size), torch.nn.ReLU()))

    # node aggregator op
    self.gnn_layers = nn.ModuleList()
    for i in range(num_layers):
        self.gnn_layers.append(NaMixedOp(hidden_size, hidden_size, self.with_linear))

    #act op
    self.act_ops = nn.ModuleList()
    for i in range(num_layers):
        self.act_ops.append(ActMixedOp())

    #readoutop
    self.readout_layers = nn.ModuleList()
    for i in range(num_layers+1):
        self.readout_layers.append(ReadoutMixedOp(hidden_size))

    #readoutop
    self.batch_norms = nn.ModuleList()
    for i in range(num_layers):
        self.batch_norms.append(nn.BatchNorm1d(hidden_size))

    #pooling ops
    self.pooling_layers = nn.ModuleList()
    for i in range(num_layers):
        self.pooling_layers.append(PoolingMixedOp(hidden_size, self.pooling_ratio[i], num_nodes=self.num_nodes))

    self.pooling_global = PoolingMixedOp(hidden_size * (num_layers + 1), args.pooling_ratio, num_nodes=self.num_nodes)
    self.pooling_trivial = Pooling_trivial(hidden_size * (num_layers + 1), args.pooling_ratio)

    #graph representation aggregator op
    self.layer6 = LaMixedOp(hidden_size, num_layers+1)
    self.layer7 = Readout_trivial()

    self.lin_output = nn.Linear(hidden_size * (num_layers+1) * 2, hidden_size)
    self.classifier = nn.Linear(hidden_size, out_dim)

    self._initialize_alphas()

  def _get_categ_mask(self, alpha):
      # log_alpha = torch.log(alpha)
      log_alpha = alpha
      u = torch.zeros_like(log_alpha).uniform_()
      softmax = torch.nn.Softmax(-1)
      one_hot = softmax((log_alpha + (-((-(u.log())).log()))) / self.temp)
      return one_hot

  def get_one_hot_alpha(self, alpha):
      one_hot_alpha = torch.zeros_like(alpha, device=alpha.device)
      idx = torch.argmax(alpha, dim=-1)

      for i in range(one_hot_alpha.size(0)):
        one_hot_alpha[i, idx[i]] = 1.0

      return one_hot_alpha

  def get_uniform_alpha(self, alpha):
      uniform_alpha = torch.full_like(alpha, 1.0 / alpha.size(-1), device=alpha.device)
      return uniform_alpha

  def forward(self, data, mode='none', graph_alpha = None):
    self.args.search_act = False
    with_linear = self.with_linear
    x, edge_index = data.x, data.edge_index
    edge_attr = getattr(data, 'edge_attr', None)
    batch = data.batch
    # edge_index, _ = remove_self_loops(edge_index)
    if edge_attr == None:
        edge_index, _ = add_self_loops(edge_index, num_nodes=x.size()[0])

    if self.args.model_type == 'darts':
        na_alphas = F.softmax(self.log_na_alphas, dim=-1)
        la_alphas = F.softmax(self.log_la_alphas, dim=-1)
        pool_alphas = F.softmax(self.log_pool_alphas, dim=-1)
        readout_alphas = F.softmax(self.log_readout_alphas, dim=-1)
        act_alphas = F.softmax(self.log_act_alphas, dim=-1)
        # print('DARTS: sampled arch in train w', self._sparse(na_alphas, act_alphas, pool_alphas, readout_alphas, la_alphas))
    else:
        na_alphas = self._get_categ_mask(self.log_na_alphas)
        # sc_alphas = self._get_categ_mask(self.log_sc_alphas)
        la_alphas = self._get_categ_mask(self.log_la_alphas)
        pool_alphas = self._get_categ_mask(self.log_pool_alphas)
        readout_alphas = self._get_categ_mask(self.log_readout_alphas)
        act_alphas = self._get_categ_mask(self.log_act_alphas)
        # print('alpha in train w:',self._arch_parameters)
        # print('sampled arch in train w', self._sparse(na_alphas, act_alphas, pool_alphas, readout_alphas, la_alphas))

    if mode == 'evaluate_single_path':
        na_alphas = self.get_one_hot_alpha(na_alphas)
        la_alphas = self.get_one_hot_alpha(la_alphas)
        pool_alphas = self.get_one_hot_alpha(pool_alphas)
        readout_alphas = self.get_one_hot_alpha(readout_alphas)
        act_alphas = self.get_one_hot_alpha(act_alphas)

    elif mode == 'mixed':
        na_alphas = self.get_uniform_alpha(na_alphas)
        pool_alphas = self.get_uniform_alpha(pool_alphas)

    elif mode == 'mads':
        # use index_select
        na_alphas = [torch.index_select(a, 0, batch) for a in graph_alpha]
        #batch_weight = torch.index_select(weights, 0, batch)

    graph_representations = []
    if not self.mol:
        x = F.elu(self.lin1(x))
    else:
        x = self.atom_encoder(x)

    edge_weights = torch.ones(edge_index.size()[1], device=edge_index.device).float()
    virtualnode_embedding = self.virtualnode_embedding(torch.zeros(batch[-1].item() + 1).to(edge_index.dtype).to(edge_index.device))

    #graph_representations.append(self.readout_layers[0](x, batch, None, readout_alphas[0]))
    gr = [x]

    for i in range(self.num_layers):
        if self.virtual:
            orix = x
            x = x + virtualnode_embedding[batch]
        if mode == "mads":
            x = self.gnn_layers[i](x, na_alphas[i], edge_index, edge_weights, edge_attr, with_linear, graph_match = True)
        else:
            x = self.gnn_layers[i](x, na_alphas[i], edge_index, edge_weights, edge_attr, with_linear)
        #print('evaluate data {}-th gnn:'.format(i), x.size(), batch.size())

        if self.args.search_act:
            x = self.act_ops[i](x, act_alphas[i])
        else:
            x = F.elu(x)
        layer_norm = nn.LayerNorm(normalized_shape=x.size(), elementwise_affine=False)
        x = layer_norm(x)
        #x = self.batch_norms[i](x)
        x = F.dropout(x, p=self.dropout, training=self.training)
        gr.append(x)

        #x, edge_index, edge_weights, batch, _ = self.pooling_layers[i](x, edge_index, edge_weights, data, batch, None, pool_alphas[i])
        #graph_representations.append(self.readout_layers[i+1](x, batch, None, readout_alphas[i+1]))

        if self.virtual and i < self.num_layers - 1:
            virtualnode_embedding_temp = global_add_pool(orix, batch) + virtualnode_embedding
            ### transform virtual nodes using MLP
            virtualnode_embedding = F.dropout(self.mlp_virtualnode_list[i](virtualnode_embedding_temp), self.dropout, training = self.training)

    #print(gr[0].size())
    gr = torch.cat(gr, 1)

    # if use pooling trivial, you should set the GNN in SAG in op_graph_classification.py
    #x, edge_index, edge_weights, batch, _ = self.pooling_trivial(gr, edge_index, edge_weights, data, batch, None)
    x, edge_index, edge_weights, batch, _ = self.pooling_global(gr, edge_index, edge_weights, data, batch, None, pool_alphas[0])

    x5 = self.layer7(x, batch)
    #x5 = self.layer6(graph_representations, la_alphas[0])

    #x5 = F.elu(self.lin_output(x5))
    x_emb = self.lin_output(x5)
    x5 = F.elu(x_emb)
    x5 = F.dropout(x5, p=self.dropout, training=self.training)
    logits = self.classifier(x5)

    if not self.mol:
        logits = F.log_softmax(logits, dim = -1)

    return logits, x_emb

  def _initialize_alphas(self):
    num_na_ops = len(NA_PRIMITIVES)
    num_la_ops = len(LA_PRIMITIVES)
    num_pool_ops = len(POOL_PRIMITIVES)
    num_readout_ops = len(READOUT_PRIMITIVES)
    num_act_ops = len(ACT_PRIMITIVES)

    """if self.args.model_type == 'mads':
        self.mhssl = MhsslBox(self.args, num_na_ops, num_pool_ops)
        self._arch_parameters = self.mhssl.it.parameters()
        return"""

    if self.args.model_type == 'darts':
        self.log_na_alphas = Variable(1e-3*torch.randn(self.num_layers, num_na_ops).cuda(), requires_grad=True)
        self.log_act_alphas = Variable(1e-3*torch.randn(self.num_layers, num_act_ops).cuda(), requires_grad=True)
        #self.log_pool_alphas = Variable(1e-3*torch.randn(self.num_layers, num_pool_ops).cuda(), requires_grad=True)
        self.log_pool_alphas = Variable(1e-3*torch.randn(1, num_pool_ops).cuda(), requires_grad=True)
        self.log_readout_alphas = Variable(1e-3*torch.randn(self.num_layers+1, num_readout_ops).cuda(), requires_grad=True)
        self.log_la_alphas = Variable(1e-3*torch.randn(1, num_la_ops).cuda(), requires_grad=True)
    else:
        self.log_na_alphas = Variable(
            torch.ones(self.num_layers, num_na_ops).normal_(self._loc_mean, self._loc_std).cuda(), requires_grad=True)
        self.log_act_alphas = Variable(
            torch.ones(self.num_layers, num_act_ops).normal_(self._loc_mean, self._loc_std).cuda(), requires_grad=True)

        #self.log_pool_alphas = Variable(
        #    torch.ones(self.num_layers, num_pool_ops).normal_(self._loc_mean, self._loc_std).cuda(), requires_grad=True)
        self.log_pool_alphas = Variable(
            torch.ones(1, num_pool_ops).normal_(self._loc_mean, self._loc_std).cuda(), requires_grad=True)
        self.log_readout_alphas = Variable(
            torch.ones(self.num_layers + 1, num_readout_ops).normal_(self._loc_mean, self._loc_std).cuda(),
            requires_grad=True)

        self.log_la_alphas = Variable(torch.ones(1, num_la_ops).normal_(self._loc_mean, self._loc_std).cuda(),
                                      requires_grad=True)

    self._arch_parameters = [
      self.log_na_alphas,
      self.log_act_alphas,
      self.log_pool_alphas,
      self.log_readout_alphas,
      self.log_la_alphas
    ]

  def arch_parameters(self):
    return self._arch_parameters

  def _sparse(self, na_weights, act_alphas, pool_alphas, readout_alphas, la_weights):
      gene = []
      na_indices = torch.argmax(na_weights, dim=-1)
      for k in na_indices:
          gene.append(NA_PRIMITIVES[k])
      #sc_indices = sc_weights.argmax(dim=-1)

      act_indices = torch.argmax(act_alphas,dim=-1)
      for k in act_indices:
          gene.append(ACT_PRIMITIVES[k])

      pooling_indices = torch.argmax(pool_alphas, dim=-1)
      for k in pooling_indices:
          gene.append(POOL_PRIMITIVES[k])
      #la_indices = la_weights.argmax(dim=-1)

      readout_indices = torch.argmax(readout_alphas,dim=-1)
      for k in readout_indices:
          gene.append(READOUT_PRIMITIVES[k])

      la_indices = torch.argmax(la_weights, dim=-1)
      for k in la_indices:
          gene.append(LA_PRIMITIVES[k])
      return '||'.join(gene)

  def genotype(self):
    gene = self._sparse(F.softmax(self.log_na_alphas, dim=-1).data.cpu(),
                        F.softmax(self.log_act_alphas, dim=-1).data.cpu(),
                        F.softmax(self.log_pool_alphas, dim=-1).data.cpu(),
                        F.softmax(self.log_readout_alphas, dim=-1).data.cpu(),
                        F.softmax(self.log_la_alphas, dim=-1).data.cpu())
    return gene

class Gads(nn.Module):
    def __init__(self, criterion, in_dim, out_dim, hidden_size, num_layers=3, dropout=0.5, epsilon=0.0, args=None, with_conv_linear=False, num_nodes=0, mol=False, virtual=False):
        super().__init__()
        num_na_ops = len(NA_PRIMITIVES)
        num_pool_ops = len(POOL_PRIMITIVES)
        self.supernet0 = GEncoder(criterion, in_dim, out_dim, args.graph_dim, 2, 0.5, epsilon, args, with_conv_linear, num_nodes, mol, virtual)
        self.supernet = Network(criterion, in_dim, out_dim, hidden_size, num_layers, dropout, epsilon, args, with_conv_linear, num_nodes, mol, virtual)
        self.ag = AG(args, num_na_ops, num_pool_ops)
        #self._arch_parameters = self.mhssl.it.parameters()
        self.explore_num = 0

    def forward(self, data):
        # mhssl
        pred0, graph_emb, sslout = self.supernet0(data, mode = 'mixed') # graph * g_d
        # ag
        graph_alpha, cosloss = self.ag(graph_emb) # layer * [graph * op]
        # final supernet
        pred, _ = self.supernet(data, mode = 'mads', graph_alpha = graph_alpha)
        return pred0, pred, cosloss, sslout

    def print_for_plot(self, data, f):
        pred0, graph_emb, sslout = self.supernet0(data, mode = 'mixed') # graph * g_d
        # ag
        graph_alpha, cosloss = self.ag(graph_emb) # layer * [graph * op]
        def writelis(li):
            for i in li:
                f.write(' ')
                f.write(str(i.item()))

        for corr, y, ge, ga1, ga2, ga3 in zip(data.corr, data.y, graph_emb, graph_alpha[0], graph_alpha[1], graph_alpha[2]):
            f.write(str(corr.item()))
            f.write(' ')
            f.write(str(y.item()))
            writelis(ge)
            writelis(ga1)
            writelis(ga2)
            writelis(ga3)
            f.write('\n')

    def mixed_train(self, data):
        pred, _ = self.supernet(data, mode = 'mixed') # graph * g_d
        return pred


import torch
from torch import nn
import torch.nn.functional as F


def _get_mask(sampled, total):
    multihot = [
        i == sampled or (isinstance(sampled, list) and i in sampled)
        for i in range(total)
    ]
    return torch.tensor(multihot, dtype=torch.bool)  # pylint: disable=not-callable


class StackedLSTMCell(nn.Module):
    def __init__(self, layers, size, bias):
        super().__init__()
        self.lstm_num_layers = layers
        self.lstm_modules = nn.ModuleList(
            [nn.LSTMCell(size, size, bias=bias) for _ in range(self.lstm_num_layers)]
        )

    def forward(self, inputs, hidden):
        prev_h, prev_c = hidden
        next_h, next_c = [], []
        for i, m in enumerate(self.lstm_modules):
            curr_h, curr_c = m(inputs, (prev_h[i], prev_c[i]))
            next_c.append(curr_c)
            next_h.append(curr_h)
            # current implementation only supports batch size equals 1,
            # but the algorithm does not necessarily have this limitation
            inputs = curr_h[-1].view(1, -1)
        return next_h, next_c


class ReinforceField:
    """
    A field with ``name``, with ``total`` choices. ``choose_one`` is true if one and only one is meant to be
    selected. Otherwise, any number of choices can be chosen.
    """

    def __init__(self, name, total, choose_one):
        self.name = name
        self.total = total
        self.choose_one = choose_one

    def __repr__(self):
        return f"ReinforceField(name={self.name}, total={self.total}, choose_one={self.choose_one})"


class ReinforceController(nn.Module):
    """
    A controller that mutates the graph with RL.

    Parameters
    ----------
    fields : list of ReinforceField
        List of fields to choose.
    lstm_size : int
        Controller LSTM hidden units.
    lstm_num_layers : int
        Number of layers for stacked LSTM.
    tanh_constant : float
        Logits will be equal to ``tanh_constant * tanh(logits)``. Don't use ``tanh`` if this value is ``None``.
    skip_target : float
        Target probability that skipconnect will appear.
    temperature : float
        Temperature constant that divides the logits.
    entropy_reduction : str
        Can be one of ``sum`` and ``mean``. How the entropy of multi-input-choice is reduced.
    """

    def __init__(
        self,
        fields,
        lstm_size=64,
        lstm_num_layers=1,
        tanh_constant=1.5,
        skip_target=0.4,
        temperature=None,
        entropy_reduction="sum",
    ):
        super(ReinforceController, self).__init__()
        self.fields = fields
        self.lstm_size = lstm_size
        self.lstm_num_layers = lstm_num_layers
        self.tanh_constant = tanh_constant
        self.temperature = temperature
        self.skip_target = skip_target

        self.lstm = StackedLSTMCell(self.lstm_num_layers, self.lstm_size, False)
        self.attn_anchor = nn.Linear(self.lstm_size, self.lstm_size, bias=False)
        self.attn_query = nn.Linear(self.lstm_size, self.lstm_size, bias=False)
        self.v_attn = nn.Linear(self.lstm_size, 1, bias=False)
        self.g_emb = nn.Parameter(torch.randn(1, self.lstm_size) * 0.1)
        self.skip_targets = nn.Parameter(
            torch.tensor(
                [1.0 - self.skip_target, self.skip_target]
            ),  # pylint: disable=not-callable
            requires_grad=False,
        )
        assert entropy_reduction in [
            "sum",
            "mean",
        ], "Entropy reduction must be one of sum and mean."
        self.entropy_reduction = torch.sum if entropy_reduction == "sum" else torch.mean
        self.cross_entropy_loss = nn.CrossEntropyLoss(reduction="none")
        self.soft = nn.ModuleDict(
            {
                field.name: nn.Linear(self.lstm_size, field.total, bias=False)
                for field in fields
            }
        )
        self.embedding = nn.ModuleDict(
            {field.name: nn.Embedding(field.total, self.lstm_size) for field in fields}
        )

    def resample(self):
        self._initialize()
        result = dict()
        for field in self.fields:
            result[field.name] = self._sample_single(field)
        return result

    def _initialize(self):
        self._inputs = self.g_emb.data
        self._c = [
            torch.zeros(
                (1, self.lstm_size),
                dtype=self._inputs.dtype,
                device=self._inputs.device,
            )
            for _ in range(self.lstm_num_layers)
        ]
        self._h = [
            torch.zeros(
                (1, self.lstm_size),
                dtype=self._inputs.dtype,
                device=self._inputs.device,
            )
            for _ in range(self.lstm_num_layers)
        ]
        self.sample_log_prob = 0
        self.sample_entropy = 0
        self.sample_skip_penalty = 0

    def _lstm_next_step(self):
        self._h, self._c = self.lstm(self._inputs, (self._h, self._c))

    def _sample_single(self, field):
        self._lstm_next_step()
        logit = self.soft[field.name](self._h[-1])
        if self.temperature is not None:
            logit /= self.temperature
        if self.tanh_constant is not None:
            logit = self.tanh_constant * torch.tanh(logit)
        if field.choose_one:
            sampled = torch.multinomial(F.softmax(logit, dim=-1), 1).view(-1)
            log_prob = self.cross_entropy_loss(logit, sampled)
            self._inputs = self.embedding[field.name](sampled)
        else:
            logit = logit.view(-1, 1)
            logit = torch.cat(
                [-logit, logit], 1
            )  # pylint: disable=invalid-unary-operand-type
            sampled = torch.multinomial(F.softmax(logit, dim=-1), 1).view(-1)
            skip_prob = torch.sigmoid(logit)
            kl = torch.sum(skip_prob * torch.log(skip_prob / self.skip_targets))
            self.sample_skip_penalty += kl
            log_prob = self.cross_entropy_loss(logit, sampled)
            sampled = sampled.nonzero().view(-1)
            if sampled.sum().item():
                self._inputs = (
                    torch.sum(self.embedding[field.name](sampled.view(-1)), 0)
                    / (1.0 + torch.sum(sampled))
                ).unsqueeze(0)
            else:
                self._inputs = torch.zeros(
                    1, self.lstm_size, device=self.embedding[field.name].weight.device
                )

        sampled = sampled.detach().numpy().tolist()
        self.sample_log_prob += self.entropy_reduction(log_prob)
        entropy = (
            log_prob * torch.exp(-log_prob)
        ).detach()  # pylint: disable=invalid-unary-operand-type
        self.sample_entropy += self.entropy_reduction(entropy)
        if len(sampled) == 1:
            sampled = sampled[0]
        return sampled



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging

import torch.nn as nn
from .mutables import Mutable, MutableScope, InputChoice
from .utils import StructuredMutableTreeNode

logger = logging.getLogger(__name__)


class BaseMutator(nn.Module):
    """
    A mutator is responsible for mutating a graph by obtaining the search space from the network and implementing
    callbacks that are called in ``forward`` in mutables.

    Parameters
    ----------
    model : nn.Module
        PyTorch model to apply mutator on.
    """

    def __init__(self, model):
        super().__init__()
        self.__dict__["model"] = model
        self._structured_mutables = self._parse_search_space(self.model)

    def _parse_search_space(
        self, module, root=None, prefix="", memo=None, nested_detection=None
    ):
        if memo is None:
            memo = set()
        if root is None:
            root = StructuredMutableTreeNode(None)
        if module not in memo:
            memo.add(module)
            if isinstance(module, Mutable):
                if nested_detection is not None:
                    raise RuntimeError(
                        "Cannot have nested search space. Error at {} in {}".format(
                            module, nested_detection
                        )
                    )
                module.name = prefix
                module.set_mutator(self)
                root = root.add_child(module)
                if not isinstance(module, MutableScope):
                    nested_detection = module
                if isinstance(module, InputChoice):
                    for k in module.choose_from:
                        if k != InputChoice.NO_KEY and k not in [
                            m.key for m in memo if isinstance(m, Mutable)
                        ]:
                            raise RuntimeError(
                                "'{}' required by '{}' not found in keys that appeared before, and is not NO_KEY.".format(
                                    k, module.key
                                )
                            )
            for name, submodule in module._modules.items():
                if submodule is None:
                    continue
                submodule_prefix = prefix + ("." if prefix else "") + name
                self._parse_search_space(
                    submodule,
                    root,
                    submodule_prefix,
                    memo=memo,
                    nested_detection=nested_detection,
                )
        return root

    @property
    def mutables(self):
        """
        A generator of all modules inheriting :class:`~nni.nas.pytorch.mutables.Mutable`.
        Modules are yielded in the order that they are defined in ``__init__``.
        For mutables with their keys appearing multiple times, only the first one will appear.
        """
        return self._structured_mutables

    @property
    def undedup_mutables(self):
        return self._structured_mutables.traverse(deduplicate=False)

    def forward(self, *inputs):
        """
        Warnings
        --------
        Don't call forward of a mutator.
        """
        raise RuntimeError("Forward is undefined for mutators.")

    def __setattr__(self, name, value):
        if name == "model":
            raise AttributeError(
                "Attribute `model` can be set at most once, and you shouldn't use `self.model = model` to "
                "include you network, as it will include all parameters in model into the mutator."
            )
        return super().__setattr__(name, value)

    def enter_mutable_scope(self, mutable_scope):
        """
        Callback when forward of a MutableScope is entered.

        Parameters
        ----------
        mutable_scope : MutableScope
            The mutable scope that is entered.
        """
        pass

    def exit_mutable_scope(self, mutable_scope):
        """
        Callback when forward of a MutableScope is exited.

        Parameters
        ----------
        mutable_scope : MutableScope
            The mutable scope that is exited.
        """
        pass

    def on_forward_layer_choice(self, mutable, *args, **kwargs):
        """
        Callbacks of forward in LayerChoice.

        Parameters
        ----------
        mutable : nni.nas.pytorch.mutables.LayerChoice
            Module whose forward is called.
        args : list of torch.Tensor
            The arguments of its forward function.
        kwargs : dict
            The keyword arguments of its forward function.

        Returns
        -------
        tuple of torch.Tensor and torch.Tensor
            Output tensor and mask.
        """
        raise NotImplementedError

    def on_forward_input_choice(self, mutable, tensor_list):
        """
        Callbacks of forward in InputChoice.

        Parameters
        ----------
        mutable : nni.nas.pytorch.mutables.InputChoice
            Mutable that is called.
        tensor_list : list of torch.Tensor
            The arguments mutable is called with.

        Returns
        -------
        tuple of torch.Tensor and torch.Tensor
            Output tensor and mask.
        """
        raise NotImplementedError

    def export(self):
        """
        Export the data of all decisions. This should output the decisions of all the mutables, so that the whole
        network can be fully determined with these decisions for further training from scratch.

        Returns
        -------
        dict
            Mappings from mutable keys to decisions.
        """
        raise NotImplementedError



from .mads import Gads

from .encoder import GEncoder

from .mads import Network
from .archgen import AG

from .genotypes import NA_PRIMITIVES, LA_PRIMITIVES, POOL_PRIMITIVES, READOUT_PRIMITIVES, ACT_PRIMITIVES


import torch

from . import BaseSpace
from .autogt_space import GraphTransformer


class AutogtSpace(BaseSpace):
    def __init__(self, args):
        super().__init__()
        # self.num_layers = args.num_layers
        # self.input_dim = input_dim  # data.x.size(-1)
        # self.output_dim = output_dim  # num_classes, dataset.num_classes
        # self.num_classes = output_dim
        # self.hidden_channels = args.hidden_channels
        # self.dropout = args.dropout
        # self.track = args.track
        # self.add_pre = add_pre
        self.args = args
        self.use_forward = True

    def build_graph(self):
        self.model = GraphTransformer(
            n_layers=self.args.n_layers,
            num_heads=self.args.num_heads,
            hidden_dim=self.args.hidden_dim,
            attention_dropout_rate=self.args.attention_dropout_rate,
            num_class=self.args.num_class,
            dropout_rate=self.args.dropout_rate,
            intput_dropout_rate=self.args.intput_dropout_rate,
            weight_decay=self.args.weight_decay,
            ffn_dim=self.args.ffn_dim,
            dataset_name=self.args.dataset_name,
            warmup_updates=self.args.warmup_updates,
            tot_updates=self.args.tot_updates,
            peak_lr=self.args.peak_lr,
            end_lr=self.args.end_lr,
            edge_type=self.args.edge_type,
            multi_hop_max_dist=self.args.multi_hop_max_dist,
            lap_dim=self.args.lap_enc_dim,
            svd_dim=self.args.svd_enc_dim,
            path=self.args.path,
        ).cuda()


    def load_model(self, path):
        self.build_graph()
        optimizer, lr_scheduler = self.model.configure_optimizers()
        scheduler = lr_scheduler['scheduler']
        info = torch.load(path)
        self.model.load_state_dict(info['model'])
        optimizer.load_state_dict(info['optimizer'])
        scheduler.load_state_dict(info['scheduler'])
        print("Load Successfully!")
        return self.model, optimizer, scheduler


    def save_model(self, optimizer, scheduler, path):
        print("Saving Model to Path: " + path)
        torch.save({'model': self.model.state_dict(), 'optimizer': optimizer.state_dict(), 'scheduler': scheduler.state_dict()}, path)
        print("Save Successfully!")



    def forward(self, data, params):
        if not self.use_forward:
            return self.prediction

        pred = self.model(data, params)
        self.current_pred = pred
        return pred

    def keep_prediction(self):
        self.prediction = self.current_pred

    def parse_model(self, selection):
        self.use_forward = False
        return self.wrap()



import json

from django.test import TestCase

from suggestion.models import Study
from suggestion.models import Trial
from suggestion.models import TrialMetric
from suggestion.early_stop_algorithm.early_stop_first_trial import (
    EarlyStopFirstTrialAlgorithm,
)


class EarlyStopFirstTrialAlgorithmTest(TestCase):
    def setUp(self):
        study_configuration_json = {
            "goal": "MAXIMIZE",
            "maxTrials": 5,
            "maxParallelTrials": 1,
            "params": [
                {
                    "parameterName": "hidden1",
                    "type": "INTEGER",
                    "minValue": 40,
                    "maxValue": 400,
                    "scalingType": "LINEAR",
                }
            ],
        }
        study_configuration = json.dumps(study_configuration_json)
        self.study = Study.create("RandomSearchStudy", study_configuration)
        trial1 = Trial.create(self.study.id, "RandomSearchTrial1")
        trial2 = Trial.create(self.study.id, "RandomSearchTrial2")
        self.trials = [trial1, trial2]
        TrialMetric.create(trial1.id, 10, 0.5)
        TrialMetric.create(trial1.id, 20, 0.6)
        TrialMetric.create(trial2.id, 10, 0.6)
        TrialMetric.create(trial2.id, 20, 0.5)

    def tearDown(self):
        pass

    # Test EarlyStopFirstTrialAlgorithm
    def test_get_early_stop_trials1(self):
        algorithm = EarlyStopFirstTrialAlgorithm()
        early_stop_trials = algorithm.get_early_stop_trials(self.trials)
        self.assertEqual(len(early_stop_trials), 1)
        self.assertEqual(early_stop_trials[0].name, "RandomSearchTrial1")



import typing as _typ
from . import BaseSpace
import torch
from .gasso_space import *
from torch.autograd import Variable
from torch import nn
import torch.nn.functional as F

GNN_LIST = [
    "gat",  # GAT with 2 heads
    "gcn",  # GCN
    "gin",  # GIN
    # "cheb",  # chebnet
    "sage",  # sage
    # "arma",
    # "sg",  # simplifying gcn
    "linear",  # skip connection
    # "skip",  # skip connection
    # "zero",  # skip connection
]


class MixedOp(nn.Module):
    def __init__(self, in_c, out_c, gnn_list=GNN_LIST):
        super(MixedOp, self).__init__()
        self._ops = nn.ModuleList()
        self.gnn_list = gnn_list
        for action in gnn_list:
            self._ops.append(gnn_map(action, in_c, out_c))

    def forward(self, x, edge_index, edge_weight, weights, selected_idx=None):
        gnn_list = self.gnn_list
        if selected_idx is None:
            fin = []
            for w, op, op_name in zip(weights, self._ops, gnn_list):
                """if op_name == "gcn":
                    w = 1.0
                else:
                    continue"""
                if edge_weight == None:
                    fin.append(w * op(x, edge_index))
                else:
                    fin.append(w * op(x, edge_index, edge_weight=edge_weight))
            return sum(fin)
            # return sum(w * op(x, edge_index) for w, op in zip(weights, self._ops))
        else:  # unchosen operations are pruned
            return self._ops[selected_idx](x, edge_index)


def Get_edges(adjs,):
    edges = []
    edges_weights = []
    for adj in adjs:
        edges.append(adj[0])
        edges_weights.append(torch.sigmoid(adj[1]))
    return edges, edges_weights


class CellWS(nn.Module):
    def __init__(self, steps, his_dim, hidden_dim, out_dim, dp, bias=True):
        super(CellWS, self).__init__()
        self.steps = steps
        self._ops = nn.ModuleList()
        self._bns = nn.ModuleList()
        self.use2 = False
        self.dp = 0.8
        for i in range(self.steps):
            if i == 0:
                inpdim = his_dim
            else:
                inpdim = hidden_dim
            if i == self.steps - 1:
                oupdim = out_dim
            else:
                oupdim = hidden_dim
            op = MixedOp(inpdim, oupdim)
            self._ops.append(op)
            self._bns.append(nn.BatchNorm1d(oupdim))

    def forward(self, x, adjs, weights):
        edges, ews = Get_edges(adjs)
        for i in range(self.steps):
            if i > 0:
                x = F.relu(x)
                x = F.dropout(x, p=self.dp, training=self.training)
            x = self._ops[i](x, edges[i], ews[i], weights[i])  # call the gcn module
        return x


class GassoSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.8,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = GNN_LIST,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim
        self.steps = layer_number
        self.dropout = dropout
        self.ops = ops
        self.use_forward = True
        self.dead_tensor = torch.nn.Parameter(
            torch.FloatTensor([1]), requires_grad=True
        )

    def build_graph(self):
        his_dim, cur_dim, hidden_dim, out_dim = (
            self.input_dim,
            self.input_dim,
            self.hidden_dim,
            self.hidden_dim,
        )
        self.cells = nn.ModuleList()
        self.cell = CellWS(
            self.steps, his_dim, hidden_dim, self.output_dim, self.dropout
        )
        his_dim = cur_dim
        cur_dim = self.steps * out_dim
        self.classifier = nn.Linear(cur_dim, self.output_dim)
        self.initialize_alphas()

    # def forward(self, x, adjs):
    def forward(self, data, adjs):
        if self.use_forward:
            # x, adjs = data.x, data.adj
            x = data.x
            x = F.dropout(x, p=self.dropout, training=self.training)

            weights = []
            for j in range(self.steps):
                weights.append(F.softmax(self.alphas_normal[j], dim=-1))

            x = self.cell(x, adjs, weights)
            x = F.log_softmax(x, dim=1)
            self.current_pred = x.detach()
            return x
        else:
            # for i in self.parameters():
            #    print(i)
            x = self.prediction + self.dead_tensor * 0
            return x

    def keep_prediction(self):
        self.prediction = self.current_pred

    """def to(self, *args, **kwargs):
        fin = super().to(*args, **kwargs)
        device = next(fin.parameters()).device
        fin.alphas_normal = [i.to(device) for i in self.alphas_normal]
        return fin"""

    def initialize_alphas(self):
        num_ops = len(self.ops)

        self.alphas_normal = []
        for i in range(self.steps):
            self.alphas_normal.append(
                Variable(1e-3 * torch.randn(num_ops), requires_grad=True)
            )

        self._arch_parameters = [self.alphas_normal]

    def arch_parameters(self):
        return self.alphas_normal

    def parse_model(self, selection):
        self.use_forward = False
        return self.wrap()



# "AutoGT: Automated Graph Transformer Architecture Search" ICLR 23'


import json
import os
import random
import time

import torch
import torch.nn.functional as F
from tqdm import tqdm, trange

from ..estimator.base import BaseEstimator
from ..space import BaseSpace
from .base import BaseNAS


class Autogt(BaseNAS):
    """
    AutoGT trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    device : str or torch.device
        The device of the whole process
    """

    def __init__(
        self,
        num_epochs=100,
        device="auto",
        disable_progress=False,
        args=None,
    ):
        super().__init__(device=device)
        self.device = device
        self.num_epochs = num_epochs
        self.disable_progress = disable_progress
        self.args = args

    def prepare(self, data):
        self.train_loader = data[0]
        self.valid_loader = data[1]
        self.test_loader = data[2]


    def train_supernet(self, optimizer, scheduler):
        self.space.train()
        total_loss = 0
        for batched_data in self.train_loader[1:]:
            optimizer.zero_grad()
            y_hat = self.space(batched_data, self.space.model.gen_params()).squeeze()
            y_gt = batched_data.y.view(-1)
            loss = F.binary_cross_entropy_with_logits(y_hat, y_gt.float())
            loss.backward()
            optimizer.step()
            total_loss += loss.double() * batched_data.y.shape[0]
            scheduler.step()
        return total_loss / self.train_loader[0]


    def train(self, optimizer, scheduler, params=None):
        self.space.train()
        total_loss = 0
        for batched_data in self.train_loader[1:]:
            optimizer.zero_grad()
            y_hat = self.space(batched_data, params).squeeze()
            y_gt = batched_data.y.view(-1)
            loss = F.binary_cross_entropy_with_logits(y_hat, y_gt.float())
            loss.backward()
            optimizer.step()
            total_loss += loss.double() * batched_data.y.shape[0]
            scheduler.step()
        return total_loss / self.train_loader[0]

    def test(self, model, data_loader, params=None):
        model.eval()
        total_correct = 0
        total_loss = 0
        for batched_data in data_loader[1:]:
            out = model(batched_data, params).squeeze()
            total_correct += int(((out > 0.5) == batched_data.y).sum())
            loss = F.binary_cross_entropy_with_logits(out, batched_data.y.view(-1).float())
            total_loss += loss.double() * batched_data.y.shape[0]
        return total_loss / data_loader[0], total_correct / data_loader[0]


    # def _infer(self, mask="train"):

    #     if mask == "train":
    #         mask = self.train_idx
    #     elif mask == "valid":
    #         mask = self.valid_idx
    #     else:
    #         mask = self.test_idx

    #     metric, loss = self.estimator.infer(self.space, self.data, self.args.arch, mask=mask)
    #     return metric, loss

    def gen_params(self, path, spa, edg, pma, cen):
        with open(path, 'r') as f:
            dic = json.load(f)
        depth = random.choice(dic['depth'])
        layers = []
        for _ in range(0, depth):
            layer = []
            hidden_in = random.choice(dic['hidden_in'])
            num_heads = random.choice(dic['num_heads'])
            att_size = random.choice(dic['att_size'])
            hidden_mid = random.choice(dic['hidden_mid'])
            ffn_size = random.choice(dic['ffn_size'])
            mask = random.choice(dic['mask'])
            layer.append((hidden_in, num_heads, att_size, hidden_mid, ffn_size, mask))
            cen = random.choice([True, False])
            eig = random.choice([True, False])
            svd = random.choice([True, False])
            layer.append((cen, eig, svd))
            spa = spa > 0
            edg = edg > 0
            pma = random.choice([True, False])
            layer.append((spa, edg, pma))
            layers.append(tuple(layer))
        return (depth, tuple(layers))

    def get_directory(self):
        directory = f'./PROTEINS/checkpoints/{self.args.dataset_name}_4/{str(self.args.seed)}/{str(self.args.data_split)}/'
        return directory

    def get_ord(self, params):
        spa = int(params[1][0][2][0])
        edg = int(params[1][0][2][1])
        pma = int(params[1][0][2][2])
        cen = int(params[1][0][1][0])
        ord = spa + (edg << 1)# + (pma << 2) + (cen << 3)
        return ord

    def gen_layer(self, dic, params):
        layer = []
        hidden_in = random.choice(dic['hidden_in'])
        num_heads = random.choice(dic['num_heads'])
        att_size = random.choice(dic['att_size'])
        hidden_mid = random.choice(dic['hidden_mid'])
        ffn_size = random.choice(dic['ffn_size'])
        mask = random.choice(dic['mask'])
        layer.append([hidden_in, num_heads, att_size, hidden_mid, ffn_size, mask])
        cen = random.choice([True, False])
        eig = random.choice([True, False])
        svd = random.choice([True, False])
        layer.append([cen, eig, svd])
        spa = params[1][0][2][0]
        edg = params[1][0][2][1]
        pma = random.choice([True, False])
        layer.append([spa, edg, pma])
        return layer

    def evolution(self, directory):
        start = time.time()
        with open(self.args.path, 'r') as f:
            dic = json.load(f)
        models = []
        for ord in range(4):
            sub_name = 'supernet_' + str(ord) + '.pt'
            models.append(self.space.load_model(directory + sub_name)[0])

        information = {}
        population = []
        candidates = []

        def is_legal(params):
            if params in information:
                return False
            info = {}
            ord = self.get_ord(params)
            _, valid_acc = self.test(models[ord], self.valid_loader, params)
            _, test_acc_ = self.test(models[ord], self.test_loader, params)
            info['valid_acc'] = valid_acc
            info['test_acc_'] = test_acc_
            # print('Top-1 Valid Accuracy = {}, Top-1 Test Accuracy = {}, Parameters = {}'.format(
            #     valid_acc, test_acc_, params))

            information[params] = info
            return True

        def get_mutation():
            # start = time.time()
            # print('Start Mutation!')
            result = []

            def random_function():
                params = random.choice(population)
                depth, layers = params
                layers = [list(list(item) for item in layer) for layer in layers]
                if random.random() < self.args.m_prob:
                    new_depth = random.choice(dic['depth'])
                    if new_depth > depth:
                        layers = layers + [self.gen_layer(dic, params) for _ in range(new_depth - depth)]
                    else:
                        layers = layers[:new_depth]
                    depth = new_depth
                for i in range(depth):
                    if random.random() < self.args.m_prob:
                        layers[i][0][0] = random.choice(dic['hidden_in'])
                    if random.random() < self.args.m_prob:
                        layers[i][0][1] = random.choice(dic['num_heads'])
                    if random.random() < self.args.m_prob:
                        layers[i][0][2] = random.choice(dic['att_size'])
                    if random.random() < self.args.m_prob:
                        layers[i][0][3] = random.choice(dic['hidden_mid'])
                    if random.random() < self.args.m_prob:
                        layers[i][0][4] = random.choice(dic['ffn_size'])
                    if random.random() < self.args.m_prob:
                        layers[i][0][5] = random.choice(dic['mask'])
                    if random.random() < self.args.m_prob:
                        layers[i][1][0] = random.choice([True, False])
                    if random.random() < self.args.m_prob:
                        layers[i][1][1] = random.choice([True, False])
                    if random.random() < self.args.m_prob:
                        layers[i][1][2] = random.choice([True, False])
                if random.random() < self.args.m_prob:
                    flag = random.choice([True, False])
                    for i in range(depth):
                        layers[i][2][0] = flag
                if random.random() < self.args.m_prob:
                    flag = random.choice([True, False])
                    for i in range(depth):
                        layers[i][2][1] = flag
                if random.random() < self.args.m_prob:
                    flag = random.choice([True, False])
                    for i in range(depth):
                        layers[i][2][2] = flag

                layers = tuple([tuple([tuple(item) for item in layer]) for layer in layers])
                result = tuple([depth, layers])
                return result

            iters = self.args.mutation_num * 10
            while len(result) < self.args.mutation_num and iters > 0:
                iters -= 1
                params = random_function()
                if not is_legal(params):
                    continue
                result.append(params)

            # end = time.time()
            # print("End Mutation! Use time: {} s".format(end - start))
            return result

        def get_hybridization():
            # start = time.time()
            # print('Start Hybridization!')
            result = []

            def random_function():
                params1 = random.choice(population)
                params2 = random.choice(population)
                iters = self.args.population_num
                while params1[0] != params2[0] and iters > 0:
                    iters -= 1
                    params1 = random.choice(population)
                    params2 = random.choice(population)
                AT_choice = random.choice([params1[1][0][2], params2[1][0][2]])
                layers = []
                for i in range(params1[0]):
                    shape_choice = []
                    for j in range(6):
                        shape_choice.append(random.choice([params1[1][i][0][j], params2[1][i][0][j]]))
                    PE_choice = []
                    for j in range(3):
                        PE_choice.append(random.choice([params1[1][i][1][j], params2[1][i][1][j]]))
                    layers.append(tuple([tuple(shape_choice), tuple(PE_choice), AT_choice]))
                params = (params1[0], tuple(layers))
                return params

            iters = 10 * self.args.hybridization_num
            while len(result) < self.args.hybridization_num and iters > 0:
                iters -= 1
                params = random_function()
                if not is_legal(params):
                    continue
                result.append(params)

            # end = time.time()
            # print("End Hybridization! Use time: {} s".format(end - start))
            return result


        # print("Evolution Started!")

        with trange(self.args.evol_epochs) as bar:
            for epoch in bar:
            # for epoch in tqdm(range(self.args.evol_epochs)):
                while len(candidates) < self.args.population_num:
                    spa = random.choice([True, False])
                    edg = random.choice([True, False])
                    pma = random.choice([True, False])
                    cen = random.choice([True, False])
                    params = self.gen_params(self.args.path, spa, edg, pma, cen)
                    if not is_legal(params):
                        continue
                    candidates.append(params)

                population += candidates
                population.sort(key=lambda x: information[x]['valid_acc'], reverse=True)
                population = population[:self.args.population_num]

                best_valid = information[population[0]]['valid_acc']
                best_test = information[population[0]]['test_acc_']
                bar.set_postfix(epoch=epoch, val_acc=best_valid, test_acc=best_test)

                # print('epoch = {} : top {} result'.format(epoch, len(population)))
                # for i, params in enumerate(population):
                #     print('No.{} Top-1 Valid Accuracy = {}, Top-1 Test Accuracy = {}, Parameters = {}'.format(
                #         i + 1, information[params]['valid_acc'], information[params]['test_acc_'], params))

                # evolution
                candidates = get_mutation() + get_hybridization()

        # end = time.time()
        # print("Evolution Ended! Use time: {} s".format(end - start))



    def fit(self):
        optimizer, lr_scheduler = self.space.model.configure_optimizers()
        scheduler = lr_scheduler['scheduler']
        best_performance = 0
        min_val_loss = float("inf")

        print('=> Phase 1: train supernet')
        for epoch in tqdm(range(self.args.split_epochs)):
            self.train_supernet(optimizer, scheduler)

        directory = self.get_directory()
        if not os.path.exists(directory):
            os.makedirs(directory)
        name = 'supernet.pt'
        self.space.save_model(optimizer, scheduler, directory + name)


        print('=> Phase 2: train subnets')
        for ord in range(4):
            spa = int((ord & 1) != 0)
            edg = int((ord & 2) != 0)
            pma = int((ord & 4) != 0)
            cen = int((ord & 8) != 0)
            _, optimizer, scheduler = self.space.load_model(directory + name)
            for epoch in tqdm(range(self.args.split_epochs, self.args.end_epochs)):
                params = self.gen_params(self.args.path, spa, edg, pma, cen)
                self.train(optimizer, scheduler, params)
            sub_name = 'supernet_' + str(ord) + '.pt'
            self.space.save_model(optimizer, scheduler, directory + sub_name)

        print('=> Phase 3: evolution')
        self.evolution(directory)



    def search(self, space: BaseSpace, data, estimator: BaseEstimator):
        self.estimator = estimator
        self.space = space.to(self.device)
        self.prepare(data)
        self.fit()
        # return space.parse_model(None)


import inspect
import sys

import torch
# from torch_geometric.utils import scatter_
import torch_scatter
special_args = [
    'edge_index', 'edge_index_i', 'edge_index_j', 'size', 'size_i', 'size_j'
]
__size_error_msg__ = ('All tensors which should get mapped to the same source '
                      'or target nodes must be of same size in dimension 0.')

is_python2 = sys.version_info[0] < 3
getargspec = inspect.getargspec if is_python2 else inspect.getfullargspec
def scatter_(name, src, index, dim=0, dim_size=None):
    assert name in ['add', 'mean', 'max']

    if name == 'max':
        op = torch.finfo if torch.is_floating_point(src) else torch.iinfo
        fill_value = op(src.dtype).min
    else:
        fill_value = 0

    op = getattr(torch_scatter, 'scatter_{}'.format(name))
    out = op(src, index, dim, None, dim_size)
    if isinstance(out, tuple):
        out = out[0]

    if name == 'max':
        out[out == fill_value] = 0

    return out

class MessagePassing(torch.nn.Module):

    def __init__(self, aggr='add', flow='source_to_target'):
        super(MessagePassing, self).__init__()

        self.aggr = aggr
        assert self.aggr in ['add', 'mean', 'max']

        self.flow = flow
        assert self.flow in ['source_to_target', 'target_to_source']

        self.__message_args__ = getargspec(self.message)[0][1:]
        self.__special_args__ = [(i, arg)
                                 for i, arg in enumerate(self.__message_args__)
                                 if arg in special_args]
        self.__message_args__ = [
            arg for arg in self.__message_args__ if arg not in special_args
        ]
        self.__update_args__ = getargspec(self.update)[0][2:]

    def propagate(self, edge_index, size=None, **kwargs):
        r"""The initial call to start propagating messages.

        Args:
            edge_index (Tensor): The indices of a general (sparse) assignment
                matrix with shape :obj:`[N, M]` (can be directed or
                undirected).
            size (list or tuple, optional): The size :obj:`[N, M]` of the
                assignment matrix. If set to :obj:`None`, the size is tried to
                get automatically inferrred. (default: :obj:`None`)
            **kwargs: Any additional data which is needed to construct messages
                and to update node embeddings.
        """

        size = [None, None] if size is None else list(size)
        assert len(size) == 2

        i, j = (0, 1) if self.flow == 'target_to_source' else (1, 0)
        ij = {"_i": i, "_j": j}

        message_args = []
        for arg in self.__message_args__:
            if arg[-2:] in ij.keys():
                tmp = kwargs.get(arg[:-2], None)
                if tmp is None:  # pragma: no cover
                    message_args.append(tmp)
                else:
                    idx = ij[arg[-2:]]
                    if isinstance(tmp, tuple) or isinstance(tmp, list):
                        assert len(tmp) == 2
                        if tmp[1 - idx] is not None:
                            if size[1 - idx] is None:
                                size[1 - idx] = tmp[1 - idx].size(0)
                            if size[1 - idx] != tmp[1 - idx].size(0):
                                raise ValueError(__size_error_msg__)
                        tmp = tmp[idx]

                    if size[idx] is None:
                        size[idx] = tmp.size(0)
                    if size[idx] != tmp.size(0):
                        raise ValueError(__size_error_msg__)

                    tmp = torch.index_select(tmp, 0, edge_index[idx])
                    message_args.append(tmp)
            else:
                message_args.append(kwargs.get(arg, None))

        size[0] = size[1] if size[0] is None else size[0]
        size[1] = size[0] if size[1] is None else size[1]

        kwargs['edge_index'] = edge_index
        kwargs['size'] = size

        for (idx, arg) in self.__special_args__:
            if arg[-2:] in ij.keys():
                message_args.insert(idx, kwargs[arg[:-2]][ij[arg[-2:]]])
            else:
                message_args.insert(idx, kwargs[arg])

        update_args = [kwargs[arg] for arg in self.__update_args__]

        out = self.message(*message_args)
        if self.aggr in ["add", "mean", "max"]:
            out = scatter_(self.aggr, out, edge_index[i], dim_size=size[i])
        else:
            pass
        out = self.update(out, *update_args)

        return out

    def message(self, x_j):  # pragma: no cover
        r"""Constructs messages in analogy to :math:`\phi_{\mathbf{\Theta}}`
        for each edge in :math:`(i,j) \in \mathcal{E}`.
        Can take any argument which was initially passed to :meth:`propagate`.
        In addition, features can be lifted to the source node :math:`i` and
        target node :math:`j` by appending :obj:`_i` or :obj:`_j` to the
        variable name, *.e.g.* :obj:`x_i` and :obj:`x_j`."""

        return x_j

    def update(self, aggr_out):  # pragma: no cover
        r"""Updates node embeddings in analogy to
        :math:`\gamma_{\mathbf{\Theta}}` for each node
        :math:`i \in \mathcal{V}`.
        Takes in the output of aggregation as first argument and any argument
        which was initially passed to :meth:`propagate`."""

        return aggr_out



import os
import logging as _logging
import typing as _typing

_backend_logger: _logging.Logger = _logging.getLogger("backend")


class _BackendConfig(_typing.Mapping[str, _typing.Any]):
    def __init__(
        self, name: str, configurations: _typing.Mapping[str, _typing.Any] = ...
    ):
        self.__name: str = name
        if configurations not in (None, Ellipsis, ...) and isinstance(
            configurations, _typing.Mapping
        ):
            self.__configurations: _typing.Mapping[str, _typing.Any] = configurations
        else:
            self.__configurations: _typing.Mapping[str, _typing.Any] = dict()

    def __str__(self) -> str:
        return self.__name

    def __getitem__(self, key: str):
        return self.__configurations[key]

    def __len__(self) -> int:
        return len(self.__configurations)

    def __iter__(self):
        return iter(self.__configurations)


class _DGLConfig(_BackendConfig):
    def __init__(self):
        super(_DGLConfig, self).__init__("dgl")


class _PyGConfig(_BackendConfig):
    def __init__(self):
        super(_PyGConfig, self).__init__("pyg")


# class _BackendConfigGenerator:
#     ...


def _generate_backend_config() -> _BackendConfig:
    def _generate_by_name(name: _typing.Optional[str] = ...) -> _BackendConfig:
        if name in (None, Ellipsis, ...) or not isinstance(name, str):
            try:
                import dgl

                return _DGLConfig()
            except ModuleNotFoundError:
                pass
            try:
                import torch_geometric

                return _PyGConfig()
            except ModuleNotFoundError:
                raise ModuleNotFoundError("Neither DGL nor PyTorch-Geometric exists")
        elif name.lower() not in ("dgl", "pyg"):
            __warning_message = " ".join(
                (
                    "The environment variable AUTOGL_BACKEND specified",
                    'but is neither "dgl" nor "pyg",',
                    "thus the environment variable is ignored",
                    "and dependent backend for AutoGL is set automatically",
                )
            )
            _backend_logger.warning(__warning_message)
            return _generate_by_name()
        elif name.lower() == "dgl":
            try:
                import dgl

                return _DGLConfig()
            except ModuleNotFoundError:
                pass
            try:
                import torch_geometric

                __warning_message: str = " ".join(
                    (
                        "The required backend DGL is not installed,",
                        "use PyTorch-Geometric instead.",
                    )
                )
                _backend_logger.warning(__warning_message)
                return _PyGConfig()
            except ModuleNotFoundError:
                raise ModuleNotFoundError("Neither DGL nor PyTorch-Geometric exists")
        elif name.lower() == "pyg":
            try:
                import torch_geometric

                return _PyGConfig()
            except ModuleNotFoundError:
                pass
            try:
                import dgl

                __warning_message: str = " ".join(
                    (
                        "The required backend PyTorch-Geometric is not installed,",
                        "use DGL instead.",
                    )
                )
                _backend_logger.warning(__warning_message)
                return _DGLConfig()
            except ModuleNotFoundError:
                raise ModuleNotFoundError("Neither DGL nor PyTorch-Geometric exists")
        else:
            return _generate_by_name()

    if "AUTOGL_BACKEND" in os.environ:
        return _generate_by_name(os.getenv("AUTOGL_BACKEND"))
    else:
        return _generate_by_name()


class _DependentBackendMetaclass(type):
    """
    Metaclass for ``DependentBackend``.
    To ensure the backend config is unique in diverse threads for multiprocessing runtime,
    the backend config is instantiated in the metaclass during interpretation phase.
    """

    def __new__(
        mcs,
        name: str,
        bases: _typing.Tuple[type, ...],
        namespace: _typing.Dict[str, _typing.Any],
    ):
        for base in bases:
            if isinstance(base, _DependentBackendMetaclass):
                strings = (
                    f"{base} is instance of Metaclass {_DependentBackendMetaclass}",
                    f"and MUST not be inherited/extended by <{name}> to construct",
                )
                raise TypeError(" ".join(strings))
        instance = super(_DependentBackendMetaclass, mcs).__new__(
            mcs, name, bases, namespace
        )
        return instance

    def __init__(
        cls,
        name: str,
        bases: _typing.Tuple[type, ...],
        namespace: _typing.Dict[str, _typing.Any],
    ):
        super(_DependentBackendMetaclass, cls).__init__(name, bases, namespace)
        cls._backend_config: _BackendConfig = _generate_backend_config()
        _backend_logger.info("Adopted backend: %s" % str(cls._backend_config))


class DependentBackend(metaclass=_DependentBackendMetaclass):
    def __new__(cls, *args, **kwargs):
        raise RuntimeError(f"The class {DependentBackend} should not be instantiated")

    @classmethod
    def get_backend_name(cls) -> str:
        return str(cls._backend_config)

    @classmethod
    def is_dgl(cls) -> bool:
        return isinstance(cls._backend_config, _DGLConfig)

    @classmethod
    def is_pyg(cls) -> bool:
        return isinstance(cls._backend_config, _PyGConfig)



# "Large-Scale Graph Neural Architecture Search" ICML 22'

import random

import torch
import torch.nn.functional as F
from tqdm import trange

from ..estimator.base import BaseEstimator
from ..space import BaseSpace
from .base import BaseNAS


class Gauss(BaseNAS):
    """
    GAUSS trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    device : str or torch.device
        The device of the whole process
    """

    def __init__(
        self,
        num_epochs=100,
        device="auto",
        disable_progress=False,
        args=None,
    ):
        super().__init__(device=device)
        self.device = device
        self.num_epochs = num_epochs
        self.disable_progress = disable_progress
        self.args = args

    def prepare(self, data):
        self.data = data.to(self.device)
        # fix random seed of train/val/test split
        random.seed(2022)
        masks = list(range(data.num_nodes))
        random.shuffle(masks)
        fold = int(data.num_nodes * 0.1)
        train_idx = masks[:fold * 6]
        val_idx = masks[fold * 6: fold * 8]
        test_idx = masks[fold * 8:]
        split_idx = {
            'train': torch.tensor(train_idx).long(),
            'valid': torch.tensor(val_idx).long(),
            'test': torch.tensor(test_idx).long()
        }

        for key in split_idx: split_idx[key] = split_idx[key].to(self.device)

        self.train_idx = split_idx['train'].to(self.device)
        self.valid_idx = split_idx['valid'].to(self.device)
        self.test_idx = split_idx['test'].to(self.device)

    def train_graph(
        self,
        optimizer,
        epoch
    ):
        self.space.train()

        optimizer.zero_grad()
        archs = self.space.sampler.samples(self.args.repeat)

        if self.args.use_curriculum:
            judgement = None
            best_acc = 0
            if epoch < self.args.warm_up:
                # min_ratio = args.min_ratio[0]
                min_ratio = epoch / self.args.epochs * (self.args.min_ratio[1] - self.args.min_ratio[0]) + self.args.min_ratio[0]
            else:
                min_ratio = epoch / self.args.epochs * (self.args.min_ratio[1] - self.args.min_ratio[0]) + self.args.min_ratio[0]
                # min_ratio = args.min_ratio[1]
                archs = sorted(archs, key=lambda x:x[1])

        for arch, score in archs:
            ratio = score if self.args.no_baseline else 1.
            out = self.space.model(self.data, arch)[self.train_idx]

            if self.args.min_clip > 0:
                ratio = max(ratio, self.args.min_clip)
            if self.args.max_clip > 0:
                ratio = min(ratio, self.args.max_clip)

            loss = F.nll_loss(out, self.data.y[self.train_idx], reduction="none") / self.args.repeat * ratio

            aggrement = (out.argmax(dim=1) == self.data.y[self.train_idx])
            cur_acc = aggrement.float().mean()
            if self.args.use_curriculum and (judgement is None or cur_acc > best_acc):
                # cal the judgement
                judgement = torch.ones_like(loss).float()
                bar = 1 / self.space.num_classes
                wrong_idxs = (~aggrement).nonzero()[:, 0] # .squeeze()
                # pass by the bar
                distributions = torch.exp(out)
                try:
                    wrong_idxs = wrong_idxs[distributions[wrong_idxs].max(dim=1)[0] > min(5 * bar, 0.7)]
                except:
                    import pdb
                    pdb.set_trace()
                sorted_idxs = distributions[wrong_idxs].max(dim=1)[0].sort(descending=True)[1][:int(self.args.max_ratio * out.size(0))]
                wrong_idxs = wrong_idxs[sorted_idxs]

                if min_ratio < 0:
                    judgement = judgement.bool()
                    judgement[wrong_idxs] = False
                else:
                    judgement[wrong_idxs] = min_ratio

                loss = loss.mean()
                best_acc = cur_acc
            else:
                if not self.args.use_curriculum:
                    loss = loss.mean()
                else:
                    if min_ratio < 0: loss = loss[judgement].mean()
                    else: loss = (loss * judgement).mean()

            loss.backward()
        optimizer.step()

        return loss.item(), cur_acc.item()

    def _infer(self, mask="train"):

        if mask == "train":
            mask = self.train_idx
        elif mask == "valid":
            mask = self.valid_idx
        else:
            mask = self.test_idx

        metric, loss = self.estimator.infer(self.space, self.data, self.args.arch, mask=mask)
        return metric, loss

    def fit(self):
        optimizer = torch.optim.Adam(self.space.model.parameters(), lr=self.args.lr, weight_decay=self.args.wd)

        # eta = self.args.eta
        best_performance = 0
        min_val_loss = float("inf")

        with trange(self.num_epochs, disable=self.disable_progress) as bar:
            for epoch in bar:
                """
                space training
                """
                self.space.train()
                # eta = (
                #     self.args.eta_max - self.args.eta
                # ) * epoch / self.num_epochs + self.args.eta
                optimizer.zero_grad()

                train_loss, train_acc = self.train_graph(
                    optimizer,
                    epoch
                )

                """
                space evaluation
                """
                self.space.eval()
                train_acc, _ = self._infer("train")
                val_acc, val_loss = self._infer("val")

                if min_val_loss > val_loss:
                    min_val_loss, best_performance = val_loss, val_acc
                    self.space.keep_prediction()

                bar.set_postfix(train_acc=train_acc["acc"], val_acc=val_acc["acc"])

        return best_performance, min_val_loss

    def search(self, space: BaseSpace, data, estimator: BaseEstimator):
        self.estimator = estimator
        self.space = space.to(self.device)
        self.prepare(data)
        perf, val_loss = self.fit()
        return space.parse_model(None)


import hyperopt

from . import register_hpo
from .advisorbase import AdvisorBaseHPOptimizer
from .suggestion.algorithm.bayesian_optimization import BayesianOptimization


@register_hpo("bayes")
class BayesAdvisor(AdvisorBaseHPOptimizer):
    """
    Bayes algorithm in `advisor` package
    See https://github.com/tobegit3hub/advisor for the package
    See .advisorbase.AdvisorBaseHPOptimizer for more information
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method = BayesianOptimization()



import torch
from torch_geometric.nn import MessagePassing
import torch.nn.functional as F
from torch_geometric.nn.conv import GINConv, GCNConv, GATConv, SAGEConv, GraphConv

from torch_scatter import scatter_mean

from ogb.graphproppred.mol_encoder import AtomEncoder,BondEncoder
from torch_geometric.utils import degree

from typing import Union, Tuple, Optional
from torch_geometric.typing import (OptPairTensor, Adj, Size, NoneType,
                                    OptTensor)
from torch import Tensor
from torch_geometric.utils import remove_self_loops, add_self_loops, softmax

from torch_sparse import SparseTensor, set_diag

class MLPConvMol(torch.nn.Module):
    def __init__(self, emb_dim):
        super().__init__()
        self.linear = torch.nn.Linear(emb_dim, emb_dim)

    def forward(self, x, edge_index, edge_attr):
        return self.linear(x)

    # def __repr__(self):
    #     return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
    #                                self.out_channels)

class GraphConvMol(GraphConv):
    def __init__(self, emb_dim):
        super().__init__(emb_dim, emb_dim)
        self.bond_encoder = BondEncoder(emb_dim = emb_dim)
        self.fuse = False

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_attr,
                edge_weight: OptTensor = None, size: Size = None) -> Tensor:
        """"""
        edge_embedding = self.bond_encoder(edge_attr)
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor, edge_weight: OptTensor)
        out = self.propagate(edge_index, x=x, edge_weight=edge_weight,
                             size=size, edge_attr=edge_embedding)
        out = self.lin_l(out)

        x_r = x[1]
        if x_r is not None:
            out += self.lin_r(x_r)

        return out

    def message(self, x_j: Tensor, edge_weight: OptTensor, edge_attr) -> Tensor:
        return F.relu(x_j + edge_attr) if edge_weight is None else edge_weight.view(-1, 1) * F.relu(x_j + edge_attr)

class SAGEConvMol(SAGEConv):
    def __init__(self, emb_dim):
        super().__init__(emb_dim, emb_dim)
        self.bond_encoder = BondEncoder(emb_dim = emb_dim)
        self.fuse = False

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_attr,
                size: Size = None) -> Tensor:
        edge_embedding = self.bond_encoder(edge_attr)
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor)
        out = self.propagate(edge_index, x=x, size=size, edge_attr=edge_embedding)
        out = self.lin_l(out)

        x_r = x[1]
        if self.root_weight and x_r is not None:
            out += self.lin_r(x_r)

        if self.normalize:
            out = F.normalize(out, p=2., dim=-1)

        return out

    def message(self, x_j, edge_attr):
        return F.relu(x_j + edge_attr)

class GATConvMol(GATConv):
    def __init__(self, emb_dim):
        super().__init__(emb_dim, emb_dim, add_self_loops = False)
        self.bond_encoder = BondEncoder(emb_dim = emb_dim)

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_attr,
                size: Size = None, return_attention_weights=None):
        # type: (Union[Tensor, OptPairTensor], Tensor, Size, NoneType) -> Tensor  # noqa
        # type: (Union[Tensor, OptPairTensor], SparseTensor, Size, NoneType) -> Tensor  # noqa
        # type: (Union[Tensor, OptPairTensor], Tensor, Size, bool) -> Tuple[Tensor, Tuple[Tensor, Tensor]]  # noqa
        # type: (Union[Tensor, OptPairTensor], SparseTensor, Size, bool) -> Tuple[Tensor, SparseTensor]  # noqa
        r"""
        Args:
            return_attention_weights (bool, optional): If set to :obj:`True`,
                will additionally return the tuple
                :obj:`(edge_index, attention_weights)`, holding the computed
                attention weights for each edge. (default: :obj:`None`)
        """
        edge_embedding = self.bond_encoder(edge_attr)
        H, C = self.heads, self.out_channels

        x_l: OptTensor = None
        x_r: OptTensor = None
        alpha_l: OptTensor = None
        alpha_r: OptTensor = None
        if isinstance(x, Tensor):
            assert x.dim() == 2, 'Static graphs not supported in `GATConv`.'
            x_l = x_r = self.lin_l(x).view(-1, H, C)
            alpha_l = (x_l * self.att_l).sum(dim=-1)
            alpha_r = (x_r * self.att_r).sum(dim=-1)
        else:
            x_l, x_r = x[0], x[1]
            assert x[0].dim() == 2, 'Static graphs not supported in `GATConv`.'
            x_l = self.lin_l(x_l).view(-1, H, C)
            alpha_l = (x_l * self.att_l).sum(dim=-1)
            if x_r is not None:
                x_r = self.lin_r(x_r).view(-1, H, C)
                alpha_r = (x_r * self.att_r).sum(dim=-1)

        assert x_l is not None
        assert alpha_l is not None

        if self.add_self_loops:
            if isinstance(edge_index, Tensor):
                num_nodes = x_l.size(0)
                if x_r is not None:
                    num_nodes = min(num_nodes, x_r.size(0))
                if size is not None:
                    num_nodes = min(size[0], size[1])
                edge_index, _ = remove_self_loops(edge_index)
                edge_index, _ = add_self_loops(edge_index, num_nodes=num_nodes)
            elif isinstance(edge_index, SparseTensor):
                edge_index = set_diag(edge_index)

        # propagate_type: (x: OptPairTensor, alpha: OptPairTensor)
        out = self.propagate(edge_index, x=(x_l, x_r),
                             alpha=(alpha_l, alpha_r), size=size, edge_attr = edge_embedding)

        alpha = self._alpha
        self._alpha = None

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if self.bias is not None:
            out += self.bias

        if isinstance(return_attention_weights, bool):
            assert alpha is not None
            if isinstance(edge_index, Tensor):
                return out, (edge_index, alpha)
            elif isinstance(edge_index, SparseTensor):
                return out, edge_index.set_value(alpha, layout='coo')
        else:
            return out

    def message(self, x_j: Tensor, alpha_j: Tensor, alpha_i: OptTensor,
                index: Tensor, ptr: OptTensor,
                size_i: Optional[int], edge_attr) -> Tensor:
        alpha = alpha_j if alpha_i is None else alpha_j + alpha_i
        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, index, ptr, size_i)
        self._alpha = alpha
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        return F.relu(x_j + edge_attr.unsqueeze(1)) * alpha.unsqueeze(-1)

### GIN convolution along the graph structure
class GINConvMol(MessagePassing):
    def __init__(self, emb_dim):
        '''
            emb_dim (int): node embedding dimensionality
        '''

        super().__init__(aggr = "add")

        self.mlp = torch.nn.Sequential(torch.nn.Linear(emb_dim, 2*emb_dim), torch.nn.BatchNorm1d(2*emb_dim), torch.nn.ReLU(), torch.nn.Linear(2*emb_dim, emb_dim))
        self.eps = torch.nn.Parameter(torch.Tensor([0]))

        self.bond_encoder = BondEncoder(emb_dim = emb_dim)

    def forward(self, x, edge_index, edge_attr):
        edge_embedding = self.bond_encoder(edge_attr)
        out = self.mlp((1 + self.eps) *x + self.propagate(edge_index, x=x, edge_attr=edge_embedding))

        return out

    def message(self, x_j, edge_attr):
        return F.relu(x_j + edge_attr)

    def update(self, aggr_out):
        return aggr_out

### GCN convolution along the graph structure
class GCNConvMol(MessagePassing):
    def __init__(self, emb_dim):
        super().__init__(aggr='add')

        self.linear = torch.nn.Linear(emb_dim, emb_dim)
        self.root_emb = torch.nn.Embedding(1, emb_dim)
        self.bond_encoder = BondEncoder(emb_dim = emb_dim)

    def forward(self, x, edge_index, edge_attr):
        x = self.linear(x)
        edge_embedding = self.bond_encoder(edge_attr)

        row, col = edge_index

        #edge_weight = torch.ones((edge_index.size(1), ), device=edge_index.device)
        deg = degree(row, x.size(0), dtype = x.dtype) + 1
        deg_inv_sqrt = deg.pow(-0.5)
        deg_inv_sqrt[deg_inv_sqrt == float('inf')] = 0

        norm = deg_inv_sqrt[row] * deg_inv_sqrt[col]

        return self.propagate(edge_index, x=x, edge_attr = edge_embedding, norm=norm) + F.relu(x + self.root_emb.weight) * 1./deg.view(-1,1)

    def message(self, x_j, edge_attr, norm):
        return norm.view(-1, 1) * F.relu(x_j + edge_attr)

    def update(self, aggr_out):
        return aggr_out



import math
import os
import random
import torch
import torch.nn.functional as F
import numpy as np
from tqdm import tqdm
import scipy
from .module.ops import uniform_sample

class BaseArchSampler(object):
    def __init__(self, space, num_layers=5):
        self.space = space
        self.num_layers = num_layers

    def sample(self):
        return uniform_sample(self.num_layers, self.space), 1.

    def fit(self, arch2score):
        return

    def train(self):
        return

    def eval(self):
        return

    def log(self, logfunc):
        return

    def save(self, folder, epoch):
        return

    def samples(self, num_sample):
        return [self.sample() for _ in range(num_sample)]

    def restart(self):
        return

class ReSampler(BaseArchSampler):
    def __init__(self, space, num_layers=5, sample_ratio=2, target="loss"):
        super().__init__(space, num_layers)
        self.sample_ratio = sample_ratio
        self.arch2score = None
        self.target = target

    def samples(self, num_samples):
        sample_time = int(num_samples * self.sample_ratio)
        # first, random sample several architectures
        archs = [uniform_sample(self.num_layers, self.space) for _ in range(sample_time)]
        # define the distribution from these archs
        losses, acc, grad = self.arch2score(archs, gradient = (self.target=="grad"))
        if self.target == "acc":
            target = acc
        elif self.target == "loss":
            target = losses
        elif self.target == "grad":
            target = grad
        elif self.target == "nloss":
            target = [-l for l in losses]
        scores = [x / sum(target) for x in target]
        selected = random.choices(list(range(sample_time)), weights=scores, k=num_samples)
        return [[archs[i], 1 / sample_time / scores[i]] for i in selected]


class RandomSampler(BaseArchSampler):
    ...

class ExpSampler(BaseArchSampler):
    def __init__(self, space, num_layers=5, num_samples=100, temprature=1.0, target="acc"):
        super().__init__(space, num_layers)
        self.num_of_ops = len(self.space)
        self.prob = np.ones((self.num_layers, self.num_of_ops)) / self.num_of_ops
        self.num_samples = num_samples
        self.temprature = temprature
        self.target = target

    def sample(self):
        arch = [None for _ in range(self.num_layers)]
        ratio = 1.0
        for layer in range(self.num_layers):
            idx = np.random.choice(self.num_of_ops, p=self.prob[layer])
            arch[layer] = self.space[idx]
            ratio *= 1.0 / self.prob[layer][idx] / self.num_of_ops
        return arch, ratio

    def evaluate(self, archs):
        scores = []
        for arch in archs:
            p = 0.0
            for i, a in enumerate(arch):
                idx = self.space.index(a)
                p += np.log(self.prob[i][idx] + 1e-3)
            scores.append(p)
        return scores

    def fit(self, arch2score):
        archs = [uniform_sample(self.num_layers, self.space) for _ in range(self.num_samples)]
        losses, acc, grad = arch2score(archs, self.target=="grad")
        if self.target == "acc":
            target = acc
        elif self.target == "loss":
            target = losses
        elif self.target == "grad":
            target = grad
        elif self.target == "nloss":
            target = [-l for l in losses]
        bin = np.zeros((self.num_layers, self.num_of_ops))
        count = np.zeros((self.num_layers, self.num_of_ops))
        for arch, value in zip(archs, target):
            for layer in range(self.num_layers):
                bin[layer][self.space.index(arch[layer])] += value
                count[layer][self.space.index(arch[layer])] += 1

        bin = bin / count
        self._cal_prob(bin)
        return

    def _cal_prob(self, bin):
        self.prob = scipy.special.softmax(bin / self.temprature, axis=1)

    def log(self, logfunc):
        logfunc(str(self.prob))

    def save(self, folder, epoch):
        import pickle
        pickle.dump(self.prob, open(os.path.join(folder, f"{epoch}-prob.pkl"), "wb"))


class BayesSampler(ExpSampler):
    def _cal_prob(self, bin):
        self.prob = bin / bin.sum(axis=1, keepdims=True)

def select(array, index):
    # print("check device")
    # print(array.device, index.device)
    eye_on_gpu = torch.eye(array.size(1), device=array.device)
    return array[eye_on_gpu[index].bool()]

class DifferentiableSampler(BaseArchSampler):
    def __init__(self, space, num_layers=5, device='cuda', repeat=10, epoch=5, lr=1e-3):
        super().__init__(space, num_layers=num_layers)
        self.param = torch.nn.Parameter(torch.ones(self.num_layers, len(self.space)))
        self.optimizer = torch.optim.Adam([self.param], lr=lr)
        self.device = device
        self.repeat = repeat
        self.epoch = epoch
        self.lr = lr

    def restart(self):
        self.param = torch.nn.Parameter(torch.ones(self.num_layers, len(self.space)))
        self.optimizer = torch.optim.Adam([self.param], lr=self.lr)

    @property
    def prob(self):
        with torch.no_grad():
            return F.softmax(self.param, dim=1).cpu().numpy()

    def evaluate(self, archs):
        scores = []
        for arch in archs:
            p = 0.0
            for i, a in enumerate(arch):
                idx = self.space.index(a)
                p += np.log(self.prob[i][idx] + 1e-3)
            scores.append(p)
        return scores

    def log(self, log):
        log(str(self.prob))

    def save(self, folder, epoch):
        torch.save(self.param, os.path.join(folder, f"{epoch}.pth"))

    @torch.no_grad()
    def sample(self):
        # sample from gumbel-softmax
        parameters = self.param.to(self.device)
        sampled = F.gumbel_softmax(parameters, dim=-1)
        arch_list = sampled.argmax(dim=1)
        prob = F.softmax(parameters, dim=-1)
        ratio = 1.0
        for a, p in zip(arch_list, prob):
            ratio *= (1 / len(self.space) / p[a]).item()
        return [self.space[i] for i in arch_list.tolist()], ratio

    def fit(self, model, data, mask):
        # sample archs and back-prop through gumbel
        model.eval()
        for _ in range(self.epoch):
            self.optimizer.zero_grad()
            for _ in range(self.repeat):
                parameters = self.param.to(self.device)
                sampled = F.gumbel_softmax(parameters, hard=True, dim=-1)
                multiplier = sampled.max(dim=1)[0]
                arch = [self.space[i] for i in sampled.argmax(dim=1).tolist()]
                out = model(data.x, data.adj_t, arch, multiplier)
                loss = F.nll_loss(out[mask], data.y.squeeze(1)[mask]) / self.repeat
                loss.backward()
            self.optimizer.step()


class RNNAgent(torch.nn.Module):
    def __init__(self, n_words, n_lengths, n_dims=100, T=1.):
        # n_words will be <begin>
        super().__init__()
        self.embedding = torch.nn.Embedding(n_words + 1, n_dims)
        self.idx_embedding = torch.nn.Embedding(n_lengths, n_dims)
        self.cell = torch.nn.GRUCell(n_dims, n_dims)
        self.out = torch.nn.Linear(n_dims, n_words, bias=False)
        self.out.weight = self.embedding.weight
        self.embedding.weight.data.mul_(0.01)
        self.n_lengths = n_lengths
        self.n_words = n_words
        self.T = T

    def sample(self, n_samples):
        # rolling up
        inp = (torch.ones(n_samples, device=next(self.parameters()).device) * self.n_words).long()
        idx = torch.ones(n_samples, device=next(self.parameters()).device)
        x = self.embedding(inp) + self.idx_embedding((idx * 0).long())
        hidden = None
        sampled_archs = []
        log_probs = []
        probs = []
        entropy = []
        for i in range(self.n_lengths):
            hidden = self.cell(x, hidden)
            # out = F.linear(hidden, self.embedding.weight[:-1]) / 10.
            out = self.out(hidden)[:,:-1] / self.T
            prob = F.softmax(out, dim=1)
            log_prob = F.log_softmax(out, dim=1)
            sampled = torch.multinomial(prob, 1)
            log_probs.append(select(log_prob, sampled[:,0]))
            probs.append(select(prob, sampled[:,0]))
            sampled_archs.append(sampled)
            entropy.append(- (log_prob * prob).sum(dim=1))
            if i < self.n_lengths - 1:
                x = self.embedding(sampled[:,0]) + self.idx_embedding((idx * (i + 1)).long())

        sampled_archs = torch.cat(sampled_archs, dim=1)
        return sampled_archs.tolist(), log_probs, probs, sum(entropy)

    @torch.no_grad()
    def evaluate(self, archs):
        self.eval()
        n_samples = len(archs)
        inp = (torch.ones(n_samples, device=next(self.parameters()).device) * self.n_words).long()
        idx = torch.ones(n_samples, device=next(self.parameters()).device)
        x = self.embedding(inp) + self.idx_embedding((idx * 0).long())
        hidden = None
        sampled_archs = []
        log_probs = []
        probs = []
        for i in range(self.n_lengths):
            hidden = self.cell(x, hidden)
            out = self.out(hidden)[:,:-1]
            log_prob = F.log_softmax(out, dim=1)
            sampled = torch.tensor([a[i] for a in archs]).long().to(log_prob.device)
            log_probs.append(select(log_prob, sampled))
            if i < self.n_lengths - 1:
                x = self.embedding(sampled) + self.idx_embedding((idx * (i + 1)).long())

        log_probs = sum(log_probs).tolist()
        return log_probs


class RLSampler(BaseArchSampler):
    def __init__(self, space, num_layers=5, device='cuda', target="acc", epochs=50, iter=2, lr=1e-3, T=1., entropy=0.):
        super().__init__(space, num_layers=num_layers)
        self.num_of_ops = len(space)
        self.sampler = RNNAgent(self.num_of_ops, num_layers, T=T).to(device)
        self.optimizer = torch.optim.Adam(self.sampler.parameters(), lr=lr)
        self.lr = lr
        self.target = target
        self.device = device
        self.epoch = epochs
        self.iter = iter
        self.T = T
        self.entropy = entropy
        self.fitted = False

    def restart(self):
        self.sampler = RNNAgent(self.num_of_ops, self.num_layers, T=self.T).to(self.device)
        self.optimizer = torch.optim.Adam(self.sampler.parameters(), lr=self.lr)
        self.fitted = False

    def train(self):
        self.sampler.train()

    def eval(self):
        self.sampler.eval()

    @torch.no_grad()
    def sample(self):
        if not self.fitted:
            return uniform_sample(self.num_layers, self.space), 1.
        self.sampler.eval()
        arch, log_p, p, _ = self.sampler.sample(1)
        ratio = 1.0
        for p_e in p:
            ratio *= 1 / self.num_of_ops / p_e.item()
        return [self.space[i] for i in arch[0]], ratio

    def fit(self, arch2score):
        self.fitted = True
        self.sampler.train()
        baseline = None
        for _ in range(self.epoch):
            self.sampler.train()
            self.optimizer.zero_grad()
            archs, log_p, _, entropy = self.sampler.sample(self.iter)
            _, accs = arch2score([[self.space[i] for i in arch] for arch in archs])
            target = accs
            if baseline is None: baseline = np.mean(target)
            else:
                baseline = baseline * 0.9 + np.mean(target) * 0.1
            reward = torch.tensor([l - baseline for l in target]).to(log_p[0].device)
            loss_REINFORCE = - sum([(reward * lp).mean() for lp in log_p])
            if self.entropy > 0:
                loss_REINFORCE -= entropy.mean() * self.entropy
            loss_REINFORCE.backward()
            self.optimizer.step()

    def save(self, folder, epoch):
        torch.save(self.sampler.state_dict(), os.path.join(folder, f"{epoch}-sampler.pth"))

    def evaluate(self, archs):
        arch2int = [[self.space.index(a) for a in arch] for arch in archs]
        return self.sampler.evaluate(arch2int)



import torch
import torch.nn as nn
import logging

from .base import BaseNAS
from ..space import (
    BaseSpace,
    replace_layer_choice,
    replace_input_choice,
    get_module_order,
    sort_replaced_module,
    PathSamplingInputChoice,
    PathSamplingLayerChoice,
    apply_fixed_architecture,
)
from tqdm import tqdm
from datetime import datetime
import numpy as np
from .rl_utils import ReinforceController, ReinforceField

LOGGER = logging.getLogger(__name__)


class GraphNasRL(BaseNAS):
    """
    RL in GraphNas.

    Parameters
    ----------
    device : torch.device
        ``torch.device("cpu")`` or ``torch.device("cuda")``.
    num_epochs : int
        Number of epochs planned for training.
    log_frequency : int
        Step count per logging.
    grad_clip : float
        Gradient clipping. Set to 0 to disable. Default: 5.
    entropy_weight : float
        Weight of sample entropy loss.
    skip_weight : float
        Weight of skip penalty loss.
    baseline_decay : float
        Decay factor of baseline. New baseline will be equal to ``baseline_decay * baseline_old + reward * (1 - baseline_decay)``.
    ctrl_lr : float
        Learning rate for RL controller.
    ctrl_steps_aggregate : int
        Number of steps that will be aggregated into one mini-batch for RL controller.
    ctrl_steps : int
        Number of mini-batches for each epoch of RL controller learning.
    ctrl_kwargs : dict
        Optional kwargs that will be passed to :class:`ReinforceController`.
    n_warmup : int
        Number of epochs for training super network.
    model_lr : float
        Learning rate for super network.
    model_wd : float
        Weight decay for super network.
    topk : int
        Number of architectures kept in training process.
    disable_progeress: boolean
        Control whether show the progress bar.
    """

    def __init__(
        self,
        device="auto",
        num_epochs=10,
        log_frequency=None,
        grad_clip=5.0,
        entropy_weight=0.0001,
        skip_weight=0,
        baseline_decay=0.95,
        ctrl_lr=0.00035,
        ctrl_steps_aggregate=100,
        ctrl_kwargs=None,
        n_warmup=100,
        model_lr=5e-3,
        model_wd=5e-4,
        topk=5,
        disable_progress=False,
        hardware_metric_limit=None,
        weight_share=True,
    ):
        super().__init__(device)
        self.num_epochs = num_epochs
        self.log_frequency = log_frequency
        self.entropy_weight = entropy_weight
        self.skip_weight = skip_weight
        self.baseline_decay = baseline_decay
        self.ctrl_steps_aggregate = ctrl_steps_aggregate
        self.grad_clip = grad_clip
        self.ctrl_kwargs = ctrl_kwargs
        self.ctrl_lr = ctrl_lr
        self.n_warmup = n_warmup
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.hist = []
        self.topk = topk
        self.disable_progress = disable_progress
        self.hardware_metric_limit = hardware_metric_limit
        self.weight_share = weight_share

    def search(self, space: BaseSpace, dset, estimator):
        self.model = space
        self.dataset = dset  # .to(self.device)
        self.estimator = estimator
        # replace choice
        self.nas_modules = []

        k2o = get_module_order(self.model)
        replace_layer_choice(self.model, PathSamplingLayerChoice, self.nas_modules)
        replace_input_choice(self.model, PathSamplingInputChoice, self.nas_modules)
        self.nas_modules = sort_replaced_module(k2o, self.nas_modules)

        # to device
        self.model = self.model.to(self.device)
        # fields
        self.nas_fields = [
            ReinforceField(
                name,
                len(module),
                isinstance(module, PathSamplingLayerChoice) or module.n_chosen == 1,
            )
            for name, module in self.nas_modules
        ]
        self.controller = ReinforceController(
            self.nas_fields,
            lstm_size=100,
            temperature=5.0,
            tanh_constant=2.5,
            **(self.ctrl_kwargs or {}),
        )
        self.ctrl_optim = torch.optim.Adam(
            self.controller.parameters(), lr=self.ctrl_lr
        )
        # train
        with tqdm(range(self.num_epochs), disable=self.disable_progress) as bar:
            for i in bar:
                l2 = self._train_controller(i)
                bar.set_postfix(reward_controller=l2)

        # selection=self.export()

        selections = [x[1] for x in self.hist]
        candidiate_accs = [-x[0] for x in self.hist]
        # print('candidiate accuracies',candidiate_accs)
        selection = self._choose_best(selections)
        arch = space.parse_model(selection)
        # print(selection,arch)
        return arch

    def _choose_best(self, selections):
        # graphnas use top 5 models, can evaluate 20 times epoch and choose the best.
        results = []
        for selection in selections:
            accs = []
            for i in tqdm(range(20), disable=self.disable_progress):
                self.arch = self.model.parse_model(selection)
                metric, loss = self._infer(mask="val")
                metric = metric["acc"]
                accs.append(metric)
            result = np.mean(accs)
            LOGGER.info(
                "selection {} \n acc {:.4f} +- {:.4f}".format(
                    selection, np.mean(accs), np.std(accs) / np.sqrt(20)
                )
            )
            results.append(result)
        best_selection = selections[np.argmax(results)]
        return best_selection

    def _train_controller(self, epoch):
        self.model.eval()
        self.controller.train()
        self.ctrl_optim.zero_grad()
        rewards = []
        baseline = None
        # diff: graph nas train 100 and derive 100 for every epoch(10 epochs), we just train 100(20 epochs). totol num of samples are same (2000)
        with tqdm(
            range(self.ctrl_steps_aggregate), disable=self.disable_progress
        ) as bar:
            for ctrl_step in bar:
                self._resample()
                metric, loss = self._infer(mask="val")
                reward = metric["acc"]

                # bar.set_postfix(acc=metric,loss=loss.item())
                LOGGER.debug(f"{self.selection}\n{metric},{loss}")
                # diff: not do reward shaping as in graphnas code
                self.hist.append([-reward, self.selection])
                if len(self.hist) > self.topk:
                    self.hist.sort(key=lambda x: x[0])
                    self.hist.pop()
                rewards.append(reward)

                if self.entropy_weight:
                    reward += (
                        self.entropy_weight * self.controller.sample_entropy.item()
                    )

                if not baseline:
                    baseline = reward
                else:
                    baseline = baseline * self.baseline_decay + reward * (
                        1 - self.baseline_decay
                    )

                loss = self.controller.sample_log_prob * (reward - baseline)
                self.ctrl_optim.zero_grad()
                loss.backward()

                self.ctrl_optim.step()

                bar.set_postfix(max_acc=max(rewards), **metric)

        LOGGER.info(
            "epoch:{}, mean rewards:{}".format(epoch, sum(rewards) / len(rewards))
        )
        return sum(rewards) / len(rewards)

    def _resample(self):
        result = self.controller.resample()
        if self.weight_share:
            for name, module in self.nas_modules:
                module.sampled = result[name]
        else:
            self.arch = self.model.parse_model(result)
        self.selection = result

    def export(self):
        self.controller.eval()
        with torch.no_grad():
            return self.controller.resample()

    def _infer(self, mask="train"):
        if self.weight_share:
            metric, loss = self.estimator.infer(self.model, self.dataset, mask=mask)
        else:
            metric, loss = self.estimator.infer(
                self.arch._model, self.dataset, mask=mask
            )
        return metric, loss



# codes in this file are reproduced from AutoAttend with some changes.
import typing as _typ
import torch
import torch.nn.functional as F
from torch import nn
from .base import BaseSpace
from .operation import act_map, gnn_map

from autogllight.utils.backend import is_dgl
from autogllight.utils.backend import BackendOperator as BK

from .autoattend_space.ops1 import OPS as OPS1
from .autoattend_space.ops2 import OPS as OPS2
from .autoattend_space.operations import agg_map

OPS = [OPS1, OPS2]


class AutoAttendNodeClassificationSpace(BaseSpace):
    """
    AutoAttend Search Space , please refer to http://proceedings.mlr.press/v139/guan21a.html for details.
    The current implementation is nc (no context weight sharing), 
    we will in future add other types of partial weight sharing proposed in the paper.

    Parameters
    ----------
    ops_type : int 
        0 or 1 , choosing from two sets of ops with index ops_type
    gnn_ops : list of str
        op names for searching, which descripts the compostion of operation pool
    act_op : str
        determine used activation function 
    agg_ops : list of str
        agg op names for searching. Only ['add','attn'] are options, as mentioned in the paper.
    """

    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.9,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops_type=0,
        gnn_ops: _typ.Sequence[_typ.Union[str, _typ.Any]] = None,
        act_op="tanh",
        head=8,
        agg_ops=["add", "attn"],
        # agg_ops=['attn'],
    ):
        super().__init__()

        assert not is_dgl(), "Now AutoAttend only support pyg"

        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.gnn_ops = gnn_ops
        self.dropout = dropout
        self.act_op = act_op
        self.ops_type = ops_type
        self.head = head
        self.agg_ops = agg_ops

    def build_graph(self):
        self.act = act_map(self.act_op)
        PRIMITIVES = list(OPS[self.ops_type].keys())
        self.gnn_map = lambda x, *args, **kwargs: OPS[self.ops_type][x](*args, **kwargs)
        self.gnn_ops = self.gnn_ops or PRIMITIVES
        self.agg_map = lambda x, *args, **kwargs: agg_map[x](*args, **kwargs)
        # self.preproc0 = nn.Linear(self.input_dim, self.input_dim)

        for layer in range(1, self.layer_number + 1):
            # stem path
            layer_stem = f"{layer-1}->{layer}"
            key = f"stem_{layer_stem}"
            self._set_layer_choice(layer_stem, key)

            # input
            key = f"in_{layer}"
            # self._set_input_choice(key,layer, choose_from=node_labels, n_chosen=1, return_mask=False)
            self._set_input_choice(
                key, layer, n_candidates=layer, n_chosen=1, return_mask=False
            )

            # agg
            key = f"agg_{layer}"
            self._set_agg_choice(layer, key=key)

        for l1 in range(1, self.layer_number + 1):
            for l2 in range(l1):
                # from l2 to l1; l2 means input layer; l1 means output layer
                layer = f"{l2}->{l1}"
                # side path
                key = f"side_{layer}"
                for i in range(2):
                    sub_key = f"{key}_{i}"
                    self._set_layer_choice(layer, sub_key)

        self._initialized = True

        # self.classifier2 = nn.Linear(self.hidden_dim, self.output_dim)

    def _set_agg_choice(self, layer, key):
        if layer == self.layer_number:
            dim = self.output_dim
        else:
            dim = self.hidden_dim
        agg_head = self.head if layer < self.layer_number else 1
        ops = [self.agg_map(op, dim, agg_head, self.dropout) for op in self.agg_ops]
        choice = self.setLayerChoice(layer, ops, key=key,)
        return choice

    def _set_layer_choice(self, layer, key):
        l1, l2 = list(map(int, layer.split("->")))
        input_dim = output_dim = self.hidden_dim

        input_dim = self.input_dim if l1 == 0 else self.hidden_dim
        output_dim = self.output_dim if l2 == self.layer_number else self.hidden_dim

        opnames = self.gnn_ops.copy()
        if input_dim != output_dim:
            for invalid_op in "ZERO IDEN".split():
                opnames.remove(invalid_op)
        concat = l2 != self.layer_number
        if self.ops_type == 0:
            ops = [
                self.gnn_map(op, input_dim, output_dim, self.dropout, concat=concat)
                for op in opnames
            ]
        elif self.ops_type == 1:
            ops = [
                self.gnn_map(
                    op, input_dim, output_dim, self.head, self.dropout, concat=concat
                )
                for op in opnames
            ]
        choice = self.setLayerChoice(l1, ops, key=key,)
        return choice

    def _set_input_choice(self, key, layer, **kwargs):
        self.setInputChoice(layer, key=key, **kwargs)

    def forward(self, data):
        x = BK.feat(data)

        def drop(x):
            x = F.dropout(x, p=self.dropout, training=self.training)
            return x

        # prev_ = self.preproc0(x)
        prev_ = x

        side_outs = {}  # {l:[side1,side2...]} collects lth layer's input
        stem_outs = []
        input = prev_
        for layer in range(1, self.layer_number + 1):
            # do layer choice for stem
            layer_stem = f"{layer-1}->{layer}"
            op = getattr(self, f"stem_{layer_stem}")
            stem_out = BK.gconv(op, data, drop(input))
            stem_out = self.act(stem_out)
            stem_outs.append(stem_out)

            # do double layer choice for sides
            for l2 in range(layer, self.layer_number + 1):
                l1 = layer - 1
                layer_side = f"{l1}->{l2}"

                side_out_list = []
                for i in range(2):
                    op = getattr(self, f"side_{layer_side}_{i}")
                    side_out = BK.gconv(op, data, drop(input))
                    side_out = self.act(side_out)  # torch.Size([2, 2708, 64])
                    side_out_list.append(side_out)
                side_out = torch.stack(side_out_list, dim=0)

                if l2 not in side_outs:
                    side_outs[l2] = []
                side_outs[l2].append(side_out)

            # select input [x1,x2,x3] from side1,side2,stem
            current_side_outs = side_outs[layer]
            side_selected = getattr(self, f"in_{layer}")(current_side_outs)
            input = [stem_outs[-1], side_selected]

            # do agg in [add , attn]
            agg = getattr(self, f"agg_{layer}")
            input = BK.gconv(agg, data, input)

        # x = self.classifier2(input)
        x = input
        return F.log_softmax(x, dim=1)



from ...suggestion.algorithm.base_hyperopt_algorithm import BaseHyperoptAlgorithm


class SimulateAnnealAlgorithm(BaseHyperoptAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    Get the new suggested trials with simulate anneal algorithm.
    """

    def __init__(self):
        super(SimulateAnnealAlgorithm, self).__init__("anneal")



import random


class AlgorithmUtil:
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    """

    @staticmethod
    def get_random_value(min_value, max_value):
        """
        Get the random value from given min and max values.

        Args:
          min_value: The min value.
          max_value: The max value.

        Return:
          The random value between min and max values.
        """
        return random.uniform(min_value, max_value)

    @staticmethod
    def get_random_int_value(min_value, max_value):
        """
        Get the random int value from given min and max values.

        Args:
          min_value: The min value.
          max_value: The max value.

        Return:
          The random int value between min and max values.
        """
        return int(random.uniform(min_value, max_value))

    @staticmethod
    def get_closest_value_in_list(input_list, objective_value):
        """
        Return the closet value for the objective value in the list.

        Args:
          input_list: Example: [-1.5, 1.5, 2.5, 4.5]
          objective_value: Example: 1.1

        Return:
          Example: 1.5
        """
        closest_value = input_list[0]
        cloeset_value_abs = abs(closest_value - objective_value)

        for current_value in input_list:
            current_value_abs = abs(current_value - objective_value)
            if current_value_abs < cloeset_value_abs:
                closest_value = current_value
                cloeset_value_abs = current_value_abs

        return closest_value

    @staticmethod
    def get_random_item_from_list(input_list):
        """
        Get the random item from given list.

        Args:
          input_list: The list.

        Return:
          The random item.
        """
        return random.choice(input_list)



from ...suggestion.algorithm.base_hyperopt_algorithm import BaseHyperoptAlgorithm


class TpeAlgorithm(BaseHyperoptAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    Get the new suggested trials with TPE algorithm.
    """

    def __init__(self):
        super(TpeAlgorithm, self).__init__("tpe")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from collections import defaultdict

import numpy as np
import torch

from .base_mutator import BaseMutator
from .mutables import LayerChoice, InputChoice
from .utils import to_list

logger = logging.getLogger(__name__)


class Mutator(BaseMutator):
    def __init__(self, model):
        super().__init__(model)
        self._cache = dict()
        self._connect_all = False

    def sample_search(self):
        """
        Override to implement this method to iterate over mutables and make decisions.

        Returns
        -------
        dict
            A mapping from key of mutables to decisions.
        """
        raise NotImplementedError

    def sample_final(self):
        """
        Override to implement this method to iterate over mutables and make decisions that is final
        for export and retraining.

        Returns
        -------
        dict
            A mapping from key of mutables to decisions.
        """
        raise NotImplementedError

    def reset(self):
        """
        Reset the mutator by call the `sample_search` to resample (for search). Stores the result in a local
        variable so that `on_forward_layer_choice` and `on_forward_input_choice` can use the decision directly.
        """
        self._cache = self.sample_search()

    def export(self):
        """
        Resample (for final) and return results.

        Returns
        -------
        dict
            A mapping from key of mutables to decisions.
        """
        sampled = self.sample_final()
        result = dict()
        for mutable in self.mutables:
            if not isinstance(mutable, (LayerChoice, InputChoice)):
                # not supported as built-in
                continue
            result[mutable.key] = self._convert_mutable_decision_to_human_readable(
                mutable, sampled.pop(mutable.key)
            )
        if sampled:
            raise ValueError(
                "Unexpected keys returned from 'sample_final()': %s",
                list(sampled.keys()),
            )
        return result

    def status(self):
        """
        Return current selection status of mutator.

        Returns
        -------
        dict
            A mapping from key of mutables to decisions. All weights (boolean type and float type)
            are converted into real number values. Numpy arrays and tensors are converted into list.
        """
        data = dict()
        for k, v in self._cache.items():
            if torch.is_tensor(v):
                v = v.detach().cpu().numpy().tolist()
            if isinstance(v, np.ndarray):
                v = v.astype(np.float32).tolist()
            data[k] = v
        return data

    def graph(self, inputs):
        """
        Return model supernet graph.

        Parameters
        ----------
        inputs: tuple of tensor
            Inputs that will be feeded into the network.

        Returns
        -------
        dict
            Containing ``node``, in Tensorboard GraphDef format.
            Additional key ``mutable`` is a map from key to list of modules.
        """
        if not torch.__version__.startswith("1.4"):
            logger.warning(
                "Graph is only tested with PyTorch 1.4. Other versions might not work."
            )
        from .graph_utils import build_graph
        from google.protobuf import json_format

        # protobuf should be installed as long as tensorboard is installed
        try:
            self._connect_all = True
            graph_def, _ = build_graph(self.model, inputs, verbose=False)
            result = json_format.MessageToDict(graph_def)
        finally:
            self._connect_all = False

        # `mutable` is to map the keys to a list of corresponding modules.
        # A key can be linked to multiple modules, use `dedup=False` to find them all.
        result["mutable"] = defaultdict(list)
        for mutable in self.mutables.traverse(deduplicate=False):
            # A module will be represent in the format of
            # [{"type": "Net", "name": ""}, {"type": "Cell", "name": "cell1"}, {"type": "Conv2d": "name": "conv"}]
            # which will be concatenated into Net/Cell[cell1]/Conv2d[conv] in frontend.
            # This format is aligned with the scope name jit gives.
            modules = mutable.name.split(".")
            path = [{"type": self.model.__class__.__name__, "name": ""}]
            m = self.model
            for module in modules:
                m = getattr(m, module)
                path.append({"type": m.__class__.__name__, "name": module})
            result["mutable"][mutable.key].append(path)
        return result

    def on_forward_layer_choice(self, mutable, *args, **kwargs):
        """
        On default, this method retrieves the decision obtained previously, and select certain operations.
        Only operations with non-zero weight will be executed. The results will be added to a list.
        Then it will reduce the list of all tensor outputs with the policy specified in `mutable.reduction`.

        Parameters
        ----------
        mutable : nni.nas.pytorch.mutables.LayerChoice
            Layer choice module.
        args : list of torch.Tensor
            Inputs
        kwargs : dict
            Inputs

        Returns
        -------
        tuple of torch.Tensor and torch.Tensor
            Output and mask.
        """
        if self._connect_all:
            return (
                self._all_connect_tensor_reduction(
                    mutable.reduction, [op(*args, **kwargs) for op in mutable]
                ),
                torch.ones(len(mutable)).bool(),
            )

        def _map_fn(op, args, kwargs):
            return op(*args, **kwargs)

        mask = self._get_decision(mutable)
        assert len(mask) == len(
            mutable
        ), "Invalid mask, expected {} to be of length {}.".format(mask, len(mutable))
        out, mask = self._select_with_mask(
            _map_fn, [(choice, args, kwargs) for choice in mutable], mask
        )
        return self._tensor_reduction(mutable.reduction, out), mask

    def on_forward_input_choice(self, mutable, tensor_list):
        """
        On default, this method retrieves the decision obtained previously, and select certain tensors.
        Then it will reduce the list of all tensor outputs with the policy specified in `mutable.reduction`.

        Parameters
        ----------
        mutable : nni.nas.pytorch.mutables.InputChoice
            Input choice module.
        tensor_list : list of torch.Tensor
            Tensor list to apply the decision on.

        Returns
        -------
        tuple of torch.Tensor and torch.Tensor
            Output and mask.
        """
        if self._connect_all:
            return (
                self._all_connect_tensor_reduction(mutable.reduction, tensor_list),
                torch.ones(mutable.n_candidates).bool(),
            )
        mask = self._get_decision(mutable)
        assert (
            len(mask) == mutable.n_candidates
        ), "Invalid mask, expected {} to be of length {}.".format(
            mask, mutable.n_candidates
        )
        out, mask = self._select_with_mask(
            lambda x: x, [(t,) for t in tensor_list], mask
        )
        return self._tensor_reduction(mutable.reduction, out), mask

    def _select_with_mask(self, map_fn, candidates, mask):
        """
        Select masked tensors and return a list of tensors.

        Parameters
        ----------
        map_fn : function
            Convert candidates to target candidates. Can be simply identity.
        candidates : list of torch.Tensor
            Tensor list to apply the decision on.
        mask : list-like object
            Can be a list, an numpy array or a tensor (recommended). Needs to
            have the same length as ``candidates``.

        Returns
        -------
        tuple of list of torch.Tensor and torch.Tensor
            Output and mask.
        """
        if (
            (isinstance(mask, list) and len(mask) >= 1 and isinstance(mask[0], bool))
            or (isinstance(mask, np.ndarray) and mask.dtype == np.bool)
            or "BoolTensor" in mask.type()
        ):
            out = [map_fn(*cand) for cand, m in zip(candidates, mask) if m]
        elif (
            (
                isinstance(mask, list)
                and len(mask) >= 1
                and isinstance(mask[0], (float, int))
            )
            or (
                isinstance(mask, np.ndarray)
                and mask.dtype in (np.float32, np.float64, np.int32, np.int64)
            )
            or "FloatTensor" in mask.type()
        ):
            out = [map_fn(*cand) * m for cand, m in zip(candidates, mask) if m]
        else:
            raise ValueError("Unrecognized mask '%s'" % mask)
        if not torch.is_tensor(mask):
            mask = torch.tensor(mask)  # pylint: disable=not-callable
        return out, mask

    def _tensor_reduction(self, reduction_type, tensor_list):
        if reduction_type == "none":
            return tensor_list
        if not tensor_list:
            return None  # empty. return None for now
        if len(tensor_list) == 1:
            return tensor_list[0]
        if reduction_type == "sum":
            return sum(tensor_list)
        if reduction_type == "mean":
            return sum(tensor_list) / len(tensor_list)
        if reduction_type == "concat":
            return torch.cat(tensor_list, dim=1)
        raise ValueError('Unrecognized reduction policy: "{}"'.format(reduction_type))

    def _all_connect_tensor_reduction(self, reduction_type, tensor_list):
        if reduction_type == "none":
            return tensor_list
        if reduction_type == "concat":
            return torch.cat(tensor_list, dim=1)
        return torch.stack(tensor_list).sum(0)

    def _get_decision(self, mutable):
        """
        By default, this method checks whether `mutable.key` is already in the decision cache,
        and returns the result without double-check.

        Parameters
        ----------
        mutable : Mutable

        Returns
        -------
        object
        """
        if mutable.key not in self._cache:
            raise ValueError('"{}" not found in decision cache.'.format(mutable.key))
        result = self._cache[mutable.key]
        logger.debug("Decision %s: %s", mutable.key, result)
        return result

    def _convert_mutable_decision_to_human_readable(self, mutable, sampled):
        # Assert the existence of mutable.key in returned architecture.
        # Also check if there is anything extra.
        multihot_list = to_list(sampled)
        converted = None
        # If it's a boolean array, we can do optimization.
        if all([t == 0 or t == 1 for t in multihot_list]):
            if isinstance(mutable, LayerChoice):
                assert len(multihot_list) == len(mutable), (
                    "Results returned from 'sample_final()' (%s: %s) either too short or too long."
                    % (mutable.key, multihot_list)
                )
                # check if all modules have different names and they indeed have names
                if len(set(mutable.names)) == len(mutable) and not all(
                    d.isdigit() for d in mutable.names
                ):
                    converted = [
                        name for i, name in enumerate(mutable.names) if multihot_list[i]
                    ]
                else:
                    converted = [
                        i for i in range(len(multihot_list)) if multihot_list[i]
                    ]
            if isinstance(mutable, InputChoice):
                assert len(multihot_list) == mutable.n_candidates, (
                    "Results returned from 'sample_final()' (%s: %s) either too short or too long."
                    % (mutable.key, multihot_list)
                )
                # check if all input candidates have different names
                if len(set(mutable.choose_from)) == mutable.n_candidates:
                    converted = [
                        name
                        for i, name in enumerate(mutable.choose_from)
                        if multihot_list[i]
                    ]
                else:
                    converted = [
                        i for i in range(len(multihot_list)) if multihot_list[i]
                    ]
        if converted is not None:
            # if only one element, then remove the bracket
            if len(converted) == 1:
                converted = converted[0]
        else:
            # do nothing
            converted = multihot_list
        return converted



import json
import numpy as np
import hyperopt
import random

from ...suggestion.models import Study
from ...suggestion.models import Trial
from ...suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm


class BaseHyperoptAlgorithm(AbstractSuggestionAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    """

    def __init__(self, algorithm_name="tpe"):

        if algorithm_name == "tpe":
            self.hyperopt_algorithm = hyperopt.tpe.suggest
        elif algorithm_name == "random_search":
            self.hyperopt_algorithm = hyperopt.rand.suggest
        elif algorithm_name == "anneal":
            self.hyperopt_algorithm = hyperopt.anneal.suggest

    def get_new_suggestions(self, study, trials=[], number=1):
        """
        Get the new suggested trials with TPE algorithm.
        """
        # Construct search space, example: {"x": hyperopt.hp.uniform('x', -10, 10), "x2": hyperopt.hp.uniform('x2', -10, 10)}
        hyperopt_search_space = {}

        # study = Study.objects.get(name=study_name)
        study_configuration_json = json.loads(study.study_configuration)
        params = study_configuration_json["params"]

        for param in params:
            param_name = param["parameterName"]

            if param["type"] == "INTEGER":
                # TODO: Support int type of search space)
                pass

            elif param["type"] == "DOUBLE":
                hyperopt_search_space[param_name] = hyperopt.hp.uniform(
                    param_name, param["minValue"], param["maxValue"]
                )

            elif param["type"] == "DISCRETE" or param["type"] == "CATEGORICAL":
                feasible_point_list = [
                    value.strip() for value in param["feasiblePoints"].split(",")
                ]
                hyperopt_search_space[param_name] = hyperopt.hp.choice(
                    param_name, feasible_point_list
                )

        # New hyperopt variables
        hyperopt_rstate = np.random.RandomState(random.randint(1, 2 ** 31 - 1))
        hyperopt_domain = hyperopt.Domain(
            None, hyperopt_search_space, pass_expr_memo_ctrl=None
        )

        hyperopt_trial_specs = []
        hyperopt_trial_results = []
        # Example: # Example: [{'tid': 0, 'idxs': {'l1_normalization': [0], 'learning_rate': [0], 'hidden2': [0], 'optimizer': [0]}, 'cmd': ('domain_attachment', 'FMinIter_Domain'), 'vals': {'l1_normalization': [0.1], 'learning_rate': [0.1], 'hidden2': [1], 'optimizer': [1]}, 'workdir': None}]
        hyperopt_trial_miscs = []
        hyperopt_trial_new_ids = []

        # Update hyperopt for trained trials with completed advisor trials
        completed_hyperopt_trials = hyperopt.Trials()

        # completed_advisor_trials = Trial.objects.filter(
        #    study_name=study_name, status="Completed")
        completed_advisor_trials = [i for i in trials if i.status == "Completed"]

        for index, advisor_trial in enumerate(completed_advisor_trials):
            # Example: {"learning_rate": 0.01, "optimizer": "ftrl"}
            parameter_values_json = json.loads(advisor_trial.parameter_values)

            # Example: {'l1_normalization': [0], 'learning_rate': [0], 'hidden2': [0], 'optimizer': [0]}
            hyperopt_trial_miscs_idxs = {}
            # Example: {'l1_normalization': [0.1], 'learning_rate': [0.1], 'hidden2': [1], 'optimizer': [1]}
            hyperopt_trial_miscs_vals = {}
            new_id = index
            hyperopt_trial_new_ids.append(new_id)
            hyperopt_trial_misc = dict(
                tid=new_id, cmd=hyperopt_domain.cmd, workdir=hyperopt_domain.workdir
            )

            for param in params:

                if param["type"] == "INTEGER":
                    pass

                elif param["type"] == "DOUBLE":
                    parameter_value = parameter_values_json[param["parameterName"]]
                    hyperopt_trial_miscs_idxs[param["parameterName"]] = [index]
                    hyperopt_trial_miscs_vals[param["parameterName"]] = [
                        parameter_value
                    ]

                elif param["type"] == "DISCRETE":
                    feasible_points_string = param["feasiblePoints"]
                    feasible_points = [
                        float(value.strip())
                        for value in feasible_points_string.split(",")
                    ]
                    parameter_value = parameter_values_json[param["parameterName"]]
                    index_of_value_in_list = feasible_points.index(parameter_value)
                    hyperopt_trial_miscs_idxs[param["parameterName"]] = [index]
                    hyperopt_trial_miscs_vals[param["parameterName"]] = [
                        index_of_value_in_list
                    ]

                elif param["type"] == "CATEGORICAL":
                    feasible_points_string = param["feasiblePoints"]
                    feasible_points = [
                        value.strip() for value in feasible_points_string.split(",")
                    ]
                    # Example: "ftrl"
                    parameter_value = parameter_values_json[param["parameterName"]]
                    index_of_value_in_list = feasible_points.index(parameter_value)
                    hyperopt_trial_miscs_idxs[param["parameterName"]] = [index]
                    hyperopt_trial_miscs_vals[param["parameterName"]] = [
                        index_of_value_in_list
                    ]

            hyperopt_trial_specs.append(None)

            hyperopt_trial_misc["idxs"] = hyperopt_trial_miscs_idxs
            hyperopt_trial_misc["vals"] = hyperopt_trial_miscs_vals
            hyperopt_trial_miscs.append(hyperopt_trial_misc)

            # TODO: Use negative objective value for loss or not

            loss_for_hyperopt = advisor_trial.objective_value
            if study_configuration_json["goal"] == "MAXIMIZE":
                # Now hyperopt only supports fmin and we need to reverse objective value for maximization
                loss_for_hyperopt = -1 * advisor_trial.objective_value

            hyperopt_trial_result = {
                "loss": loss_for_hyperopt,
                "status": hyperopt.STATUS_OK,
            }
            hyperopt_trial_results.append(hyperopt_trial_result)

        if len(completed_advisor_trials) > 0:
            # Example: {'refresh_time': datetime.datetime(2018, 9, 18, 12, 6, 41, 922000), 'book_time': datetime.datetime(2018, 9, 18, 12, 6, 41, 922000), 'misc': {'tid': 0, 'idxs': {'x2': [0], 'x': [0]}, 'cmd': ('domain_attachment', 'FMinIter_Domain'), 'vals': {'x2': [-8.137088361136204], 'x': [-4.849028446711832]}, 'workdir': None}, 'state': 2, 'tid': 0, 'exp_key': None, 'version': 0, 'result': {'status': 'ok', 'loss': 14.849028446711833}, 'owner': None, 'spec': None}
            hyperopt_trials = completed_hyperopt_trials.new_trial_docs(
                hyperopt_trial_new_ids,
                hyperopt_trial_specs,
                hyperopt_trial_results,
                hyperopt_trial_miscs,
            )
            for current_hyperopt_trials in hyperopt_trials:
                current_hyperopt_trials["state"] = hyperopt.JOB_STATE_DONE

            completed_hyperopt_trials.insert_trial_docs(hyperopt_trials)
            completed_hyperopt_trials.refresh()

        rval = hyperopt.FMinIter(
            self.hyperopt_algorithm,
            hyperopt_domain,
            completed_hyperopt_trials,
            max_evals=-1,
            rstate=hyperopt_rstate,
            verbose=0,
        )
        rval.catch_eval_exceptions = False

        new_ids = rval.trials.new_trial_ids(number)

        rval.trials.refresh()

        random_state = rval.rstate.randint(2 ** 31 - 1)
        new_trials = self.hyperopt_algorithm(
            new_ids, rval.domain, completed_hyperopt_trials, random_state
        )
        rval.trials.refresh()

        # Construct return advisor trials from new hyperopt trials
        return_trial_list = []

        for i in range(number):

            # Example: {u'hidden2': [2], u'learning_rate': [0.04633366105812467], u'l1_normalization': [0.16858448611765364], u'optimizer': [3]}
            vals = new_trials[i]["misc"]["vals"]

            new_advisor_trial = Trial.create(study.name, "TpeTrial")
            parameter_values_json = {}

            for param in params:

                if param["type"] == "INTEGER":
                    pass

                elif param["type"] == "DOUBLE":
                    suggest_value = vals[param["parameterName"]][0]
                    parameter_values_json[param["parameterName"]] = suggest_value

                elif param["type"] == "DISCRETE":
                    feasible_point_list = [
                        float(value.strip())
                        for value in param["feasiblePoints"].split(",")
                    ]
                    suggest_index = vals[param["parameterName"]][0]
                    suggest_value = feasible_point_list[suggest_index]

                elif param["type"] == "CATEGORICAL":
                    feasible_point_list = [
                        value.strip() for value in param["feasiblePoints"].split(",")
                    ]
                    suggest_index = vals[param["parameterName"]][0]
                    suggest_value = feasible_point_list[suggest_index]

                parameter_values_json[param["parameterName"]] = suggest_value

            new_advisor_trial.parameter_values = json.dumps(parameter_values_json)
            # new_advisor_trial.save()
            return_trial_list.append(new_advisor_trial)

        return return_trial_list



from abc import abstractmethod
import torch.nn as nn
import json
from copy import deepcopy
import torch
from .nni import (
    apply_fixed_architecture,
    OrderedLayerChoice,
    OrderedInputChoice,
)


class BoxModel(nn.Module):
    """
    The box wrapping a space, can be passed to later procedure or trainer

    Parameters
    ----------
    space_model : BaseSpace
        The space which should be wrapped
    device : str or torch.device
        The device to place the model
    """

    def __init__(self, space_model, *args, **kwargs):
        super().__init__()
        self.init = True
        self.space = []
        self.hyperparams = {}
        self._model = space_model
        self.num_features = self._model.input_dim
        self.num_classes = self._model.output_dim
        self.params = {"num_class": self.num_classes, "features_num": self.num_features}
        self.selection = None

    def fix(self, selection):
        """
        To fix self._model with a selection

        Parameters
        ----------
        selection : dict
            A seletion indicating the choices of mutables
        """
        self.selection = selection
        self._model.instantiate()
        apply_fixed_architecture(self._model, selection, verbose=False)
        return self

    def forward(self, *args, **kwargs):
        return self._model(*args, **kwargs)

    def __repr__(self) -> str:
        return str({"model": self.model, "selection": self.selection})


class BaseSpace(nn.Module):
    """
    Base space class of NAS module. Defining space containing all models.
    Please use mutables to define your whole space. Refer to
    `https://nni.readthedocs.io/en/stable/NAS/WriteSearchSpace.html`
    for detailed information.

    Parameters
    ----------
    init: bool
        Whether to initialize the whole space. Default: `False`
    """

    def __init__(self):
        super().__init__()
        self._initialized = False
        self._default_key = 0

    @abstractmethod
    def forward(self, *args, **kwargs):
        """
        Define the forward pass of space model
        """
        raise NotImplementedError()

    def instantiate(self, **kwargs):
        """
        Instantiate modules in the space
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.build_graph()
        self._initialized = True

    @abstractmethod
    def build_graph(self):
        """
        Instantiate the operation graph in the space
        """
        raise NotImplementedError()

    def getOriKey(self, key):
        orikey = key
        if orikey == None:
            key = f"default_key_{self._default_key}"
            self._default_key += 1
            orikey = key
        return orikey

    def setLayerChoice(
        self, order, op_candidates, reduction="sum", return_mask=False, key=None
    ):
        """
        Give a unique key if not given
        """
        key = self.getOriKey(key)
        layer = OrderedLayerChoice(order, op_candidates, reduction, return_mask, key)
        setattr(self, key, layer)
        return layer

    def setInputChoice(
        self,
        order,
        n_candidates=None,
        choose_from=None,
        n_chosen=None,
        reduction="sum",
        return_mask=False,
        key=None,
    ):
        """
        Give a unique key if not given
        """
        key = self.getOriKey(key)
        layer = OrderedInputChoice(
            order, n_candidates, choose_from, n_chosen, reduction, return_mask, key
        )
        setattr(self, key, layer)
        return layer

    def wrap(self):
        return BoxModel(self)

    def parse_model(self, selection):
        """Get the fixed model from the selection
        Usage: the fixed model can be obtained by boxmodel._model 
        Warning : this method will randomize the learnable parameters in the model, as the model is re-instantiated.
        """
        boxmodel = self.wrap().fix(selection)
        return boxmodel


"""
BoxModel is the space itself, but without replacing the operation choices.
Therefore, the choices in BoxModel is subclasses of Mutables, which can be collected in functions like apply_fixed_architecture.
In this way, the fixed architecture "BoxModel" will get fixed operations, while the original space will not be changed.
Moreover, the fixed architecture "BoxModel" will not have multiple operations mixed together as in DARTS.  
"""



from django.test import TestCase

from suggestion.early_stop_algorithm.abstract_early_stop import (
    AbstractEarlyStopAlgorithm,
)


class RandomSearchAlgorithmTest(TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init(self):
        pass

        # abstractEarlyStopAlgorithm = AbstractEarlyStopAlgorithm()
        # self.assertEqual(abstractEarlyStopAlgorithm.__class__, AbstractEarlyStopAlgorithm)



"""
Base class for algorithm
"""
import torch
from abc import abstractmethod
from autogllight.utils import get_device


class BaseNAS:
    """
    Base NAS algorithm class

    Parameters
    ----------
    device: str or torch.device
        The device of the whole process
    """

    def __init__(self, device="auto") -> None:
        self.device = get_device(device)
        self.selection = None

    def to(self, device):
        """
        Change the device of the whole NAS search process

        Parameters
        ----------
        device: str or torch.device
        """
        self.device = get_device(device)

    @abstractmethod
    def search(self, space, dataset, estimator, return_model=True):
        """
        The search process of NAS.

        Parameters
        ----------
        space : autogllight.nas.space.BaseSpace
            The search space. Constructed following nni.
        dataset : any
            Dataset to perform search on.
        estimator : autoglight.nas.estimator.BaseEstimator
            The estimator to compute loss & metrics.

        Returns
        -------
        model: autogllight.nas.space.BaseSpace || dict
            The searched model (return_model=True) or the selection (return_model=False).
        """
        raise NotImplementedError()

    def get_selection(self):
        return self.selection



import numpy as np
import random
import os
import torch


def set_seed(seed):
    np.random.seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.enabled = False
    torch.backends.cudnn.benchmark = False


import torch
from typing import Union


def get_device(device: Union[str, torch.device]):
    """
    Get device of passed argument. Will return a torch.device based on passed arguments.
    Can parse auto, cpu, gpu, cpu:x, gpu:x, etc. If auto is given, will automatically find
    available devices.


    Parameters
    ----------
    device: ``str`` or ``torch.device``
        The device to parse. If ``auto`` if given, will determine automatically.

    Returns
    -------
    device: ``torch.device``
        The parsed device.
    """
    assert isinstance(
        device, (str, torch.device)
    ), "Only support device of str or torch.device, get {} instead".format(device)
    if device == "auto":
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    return torch.device(device)



# tackle different backend
from .depend import DependentBackend


def is_dgl():
    return DependentBackend.is_dgl()


class BackendOperator:
    """Operators based on different backends"""

    @staticmethod
    def mask(data, mask):
        return data.ndata[f"{mask}_mask"] if is_dgl() else data[f"{mask}_mask"]

    @staticmethod
    def label(data):
        return data.ndata["label"] if is_dgl() else data.y

    @staticmethod
    def feat(data):
        return data.ndata["feat"] if is_dgl() else data.x

    @staticmethod
    def gconv(op, data, feat):
        return op(data, feat) if is_dgl() else op(feat, data.edge_index)


bk_mask = BackendOperator.mask
bk_label = BackendOperator.label
bk_feat = BackendOperator.feat
bk_gconv = BackendOperator.gconv



import torch.nn.functional as F
from ..space import BaseSpace
from .base import BaseEstimator
from autogllight.utils.evaluation import Acc
from autogllight.utils.backend.op import *


class TrainScratchEstimator(BaseEstimator):
    """
    Train scratch estimator.

    Train the model to get estimations.

    Parameters
    ----------
    trainer : str
        The trainer to train the model
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
    """

    def __init__(self, trainer, evaluation=[Acc()]):
        super().__init__(None, evaluation)
        self.trainer = trainer
        self.evaluation = evaluation        

    def infer(self, model: BaseSpace, dataset, mask="train", *args, **kwargs):
        metrics, loss = self.trainer(model, dataset, mask, self.evaluation, *args, **kwargs)
        return metrics, loss
